/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package actor;

import java.util.HashMap;
import java.util.Map;
import logging.Logger;
import org.w3c.dom.Node;

/**
 * Klasse eine Actors.
 * @author nico3000
 */
public final class Actor {    
    /**
     * Die Komponenten des Actors.
     */
    private final Map<Integer, ActorComponent> components = new HashMap<>();
    /**
     * Die ID des Actors.
     */
    private final int id;
    /**
     * Der Name des Actors.
     */
    private final String name;
    /**
     * Der Name des Elternactors.
     */
    private final String parentActorName;
    
    /**
     * Konstruktor mit ID.
     * @param id Die ID des Actors
     * @param name Der Name des Actors
     * @param parentActorName Der Name des Elternactors
     */
    Actor(int id, String name, String parentActorName) {
        this.id = id;
        if(name != null && !name.trim().isEmpty()) {
            this.name = name.trim();
        } else {
            this.name = "_actor" + id;
        }
        if(parentActorName != null && !parentActorName.trim().isEmpty()) {
            this.parentActorName = parentActorName.trim();
        } else {
            this.parentActorName = null;
        }
    }

    /**
     * Getter fuer die ID.
     * @return Die ID
     */
    public int getId() {
        return this.id;
    }

    /**
     * Liefert den Namen des Actors.
     * @return Der Name
     */
    public String getName() {
        return name;
    }

    public String getParentActorName() {
        return parentActorName;
    }
    
    /**
     * Initialisiert den Actor und alle Komponenten.
     * @param initialData Die Initialisierungsdaten des Actors
     * @return <tt>true</tt>, falls erfolgreich
     */
    boolean init(Node initialData, Node overrideData) {
        if(!initialData.getNodeName().equals("Actor")) {
            Logger.INSTANCE.error("actor data does not start with node Actor");
            return false;
        } else {
            return true;
        }
    }
    
    /**
     * Wird direkt nach der Initialisierung aufgerufen.
     */
    void postInit() {
        for(ActorComponent component : this.components.values()) {
            component.postInit();
        }
    }
    
    /**
     * Wird jeden Frame aufgerufen.
     * @param deltaMillis Die Dauer des Frames in Millieskunden (Gametimer)
     */
    public void update(long deltaMillis) {
        for(ActorComponent component : this.components.values()) {
            component.update(deltaMillis);
        }
    }
    
    /**
     * Getter fuer eine bestimmte Komponente.
     * @param id Die ID der Komponentenklasse
     * @return Die Komponente, oder <tt>null</tt>
     */
    public ActorComponent getComponent(int id) {
        return this.components.get(id);
    }
    
    /**
     * Fuegt dem Actor eine Komponente hinzu.
     * @param component Die Komponente
     */
    void addComponent(ActorComponent component) {
        this.components.put(component.getComponentId(), component);
    }
    
    public void destroy() {
        for(ActorComponent component : this.components.values()) {
            component.destroy();
        }
    }
}
