/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package config;

import java.awt.Color;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.logging.Level;
import logging.Logger;

/**
 *
 * @author nico3000
 */
public class Config {
    /**
     * Name der ini-Datei.
     */
    private String filename;
    /**
     * InputStream der Datei mit den Standardwerten.
     */
    private final InputStream defaultConfigStream;
    /**
     * Sektionen der ini-Datei mit den zugehoerigen Keys.
     */
    private final Map<String, Set<String>> sections = new LinkedHashMap<>();
    /**
     * Keys der ini-Datei und die momentanen Werte.
     */
    private final Map<String, String> values = new HashMap<>();
    /**
     * Aktuelle Sektion waehrend der Initialisierung.
     */
    private String currentSection = null;
    /**
     * Keys, die keiner Sektion zugeordnet sind.
     */
    private final Set<String> unknownSection = new LinkedHashSet<>();
    
    /**
     * Konstruktor.
     * @param filename Name der ini-Datei
     * @param defaultConfigStream Stream zur Datei mit den Standardwerten
     */
    public Config(String filename, InputStream defaultConfigStream) {
        this.filename = filename;
        this.defaultConfigStream = defaultConfigStream;
    }
    
    /**
     * Liest die Default ini-Datei ein und ergaenzt sie um die eigene ini-Datei.
     * @return <tt>true</tt>, falls erfolgreich; <tt>false</tt> sonst
     */
    public boolean init() {
        if(!this.values.isEmpty()) {
            Logger.INSTANCE.warning("Config file " + this.filename + " is already initialized.");
            return false;
        } else {
            // parsing default config file
            InputStream input = this.defaultConfigStream;
            if(input == null) {
                Logger.INSTANCE.warning("No default file for " + this.filename + ".");
            } else {
                this.parseIni(new BufferedReader(new InputStreamReader(input)));
            }
            this.currentSection = null;
            // parsing custom config file
            File configFile = new File(this.filename);
            if(!configFile.exists()) {
                Logger.INSTANCE.info("Config file " + this.filename + " does not exist, so it will be created.");
                this.save();
            } else {
                try {
                    this.parseIni(new BufferedReader(new FileReader(configFile)));
                } catch (FileNotFoundException ex) {
                    Logger.INSTANCE.error(ex.getMessage());
                }
            }
            return true;
        }
    }
    
    /**
     * List eine ini-Datei von einem Reader.
     * @param iniReader Der Reader
     */
    private void parseIni(BufferedReader iniReader) {
        try {
            String line;
            while((line = iniReader.readLine()) != null) {
                this.parseLine(line);
            }
        } catch (IOException ex) {
            Logger.INSTANCE.warning("Error while reading default file for " + this.filename + ".");
        } finally {
            try {
                iniReader.close();
            } catch (IOException ex) {
                Logger.INSTANCE.error(ex.getMessage());
            }
        }
    }
    
    /**
     * Speichert die aktuelle Konfiguration in die ini-Datei.
     */
    public void save() {
        BufferedWriter writer = null;
        try {
            boolean written = false;
            writer = new BufferedWriter(new FileWriter(this.filename));
            for(String section : this.sections.keySet()) {
                if(section != null) {
                    writer.write(String.format("[%s]\n", section));
                    written = true;
                } else if(!this.sections.get(section).isEmpty()) {
                    writer.write("[UNKNOWN]\n");
                    written = true;
                }
                for(String option : this.sections.get(section)) {
                    writer.write(String.format("%s=%s\n", option, this.values.get(option)));
                    written = true;
                }
                if(written) {
                    writer.write("\n");
                }
            }
        } catch (IOException ex) {
            java.util.logging.Logger.getLogger(Config.class.getName()).log(Level.SEVERE, null, ex);
        } finally {
            if(writer != null) {
                try {
                    writer.close();
                } catch (IOException ex) {
                    java.util.logging.Logger.getLogger(Config.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
        }
    }
    
    /**
     * Interpretiert eine Zeile einer ini-Datei.
     * @param line Die Zeile
     */
    private void parseLine(String line) {
        line = line.trim();
        if(!line.isEmpty() && !line.startsWith("#")) {
            if(line.startsWith("[") && line.endsWith("]")) {
                this.currentSection = line.substring(1, line.length() - 1).trim();
                if(!this.sections.containsKey(this.currentSection)) {
                    this.sections.put(this.currentSection, new LinkedHashSet<String>());
                }
            } else {
                String split[] = line.split("=", 2);
                if(split.length == 1) {
                    Logger.INSTANCE.warning("Malformed line " + line + " in " + this.filename + " ignored (must be \"<key>=<value>\" with non-empty strings <key> and <value>).");
                } else {
                    String key = split[0].trim();
                    String value = split[1].trim();
                    String section = this.getSectionOf(key);
                    if(section == null && !this.unknownSection.contains(key)) { // key is completely unknown
                        if(this.currentSection != null) {
                            this.sections.get(this.currentSection).add(key);
                        } else {
                            this.unknownSection.add(key);
                        }
                    } else if(section == null) {                                // key is known but has no section
                        if(this.currentSection != null) {
                            this.unknownSection.remove(key);
                            this.sections.get(this.currentSection).add(key);                            
                        }
                    } else {                                                    // key has section
                        if(this.currentSection != null && !section.equals(this.currentSection)) {
                            Logger.INSTANCE.warning("Assigning key \"" + key + "\" from  section \"" + section + "\" to section \"" + this.currentSection + "\" forbidden.");
                        }
                    }
                    this.values.put(key, value);
                }
            }
        }
    }
    
    /**
     * Liefert die Sektion eines Keys, falls sie bekannt ist.
     * @param key Der Key
     * @return Die Sektion des Keys oder <tt>null</tt>, falls nicht bekannt
     */
    private String getSectionOf(String key) {
        for(String section : this.sections.keySet()) {
            if(this.sections.get(section).contains(key)) {
                return section;
            }
        }
        return null;
    }
    
    /**
     * Liefert den Wert eines Keys als String.
     * @param key Der Key
     * @return Der Wert oder <tt>null</tt>, falls der Key nicht exsitiert
     */
    public String getAsString(String key) {
        String value = this.values.get(key);
        if(value == null) {
            Logger.INSTANCE.error("Unknown key \"" + key + "\"!");
        }
        return value;
    }
    
    /**
     * Liefert den Wert eines Keys als Integer.
     * @param key Der Key
     * @return Der Wert oder <tt>0</tt>, falls der Key nicht exsitiert
     */
    public int getAsInt(String key) {
        if(!key.startsWith("i")) {
            Logger.INSTANCE.warning("Trying to get int value from the not-int-marked key \"" + key + "\".");
        }
        int value = 0;
        try {
            value = Integer.parseInt(this.getAsString(key));
        } catch(NumberFormatException ex) {
            Logger.INSTANCE.error("The key \"" + key + "\" is no int!");
        }
        return value;
    }
    
    /**
     * Liefert den Wert eines Keys als Boolean.
     * @param key Der Key
     * @return Der Wert oder <tt>false</tt>, falls der Key nicht exsitiert
     */
    public boolean getAsBoolean(String key) {
        if(!key.startsWith("b")) {
            Logger.INSTANCE.warning("Trying to get boolean value from the not-boolean-marked key \"" + key + "\".");
        }
        return Boolean.parseBoolean(this.getAsString(key));
    }
    
    /**
     * Liefert den Wert eines Keys als Float.
     * @param key Der Key
     * @return Der Wert oder <tt>0.0f</tt> wenn der Key nicht exsitiert
     */
    public double getAsFloat(String key) {
        if(!key.startsWith("f")) {
            Logger.INSTANCE.warning("Trying to get float value from the not-float-marked key \"" + key + "\".");
        }
        Float value = 0.0f;
        try {
            value = Float.parseFloat(this.getAsString(key));
        } catch(NumberFormatException ex) {
            Logger.INSTANCE.error("The key \"" + key + "\" is no float!");
        }
        return value;
    }
    
    /**
     * Liefert den Wert eines Keys als Farbe.
     * @param key Der key
     * @param defaultColor Wird zurueckgegeben, falls der Key nicht exsitiert
     *                     oder keine gueltige Farbe enthaelt.
     * @return Der Wert oder <tt>null</tt>, falls der Key nicht exsitiert, oder
     *         keine gueltige Farbe enthaelt.
     */
    public Color getAsColor(String key, Color defaultColor) {
        if(!key.startsWith("c")) {
            Logger.INSTANCE.warning("Trying to get color value from the not-color-marked key \"" + key + "\".");
        }
        String color = this.getAsString(key);
        if(color == null) {
            return defaultColor;
        }
        StringTokenizer tokenizer = new StringTokenizer(color);
        if(tokenizer.countTokens() < 3 || tokenizer.countTokens() > 4) {
            Logger.INSTANCE.error("Invalid color for key \"" + key + "\".");
            return null;
        }
        try {
            int red = Integer.parseInt(tokenizer.nextToken());
            int green = Integer.parseInt(tokenizer.nextToken());
            int blue = Integer.parseInt(tokenizer.nextToken());
            int alpha = tokenizer.hasMoreTokens() ? Integer.parseInt(tokenizer.nextToken()) : 255;
            return new Color(red, green, blue, alpha);
        } catch (NumberFormatException ex) {
            Logger.INSTANCE.error("Invalid color for key \"" + key + "\".");
            return defaultColor;
        }        
    }
    
    /**
     * Liefert alle Keys einer Sektion der Configdatei.
     * @param sectionName Der Name der Sektion
     * @return Menge aller Keys der Sektion
     */
    public Set<String> getSection(String sectionName) {
        Set<String> set = this.sections.get(sectionName);
        return set == null ? new LinkedHashSet<String>() : Collections.unmodifiableSet(set);
    }
}
