/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package graphics.texture;

import java.awt.image.BufferedImage;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.util.HashMap;
import java.util.Map;
import logging.Logger;
import main.GameApp;
import org.lwjgl.BufferUtils;
import resource.ResHandle;
import resource.Resource;
import resource.loader.ImageExtraData;
import util.ImageUtil;

/**
 * Der TextureLoader ist die zentrale Klasse, um Texturen zu laden oder zu
 * erstellen.
 * @author nico3000
 */
public class TextureLoader {
    /**
     * Die Standard 2D Textur.
     */
    public static final Texture DEFAULT_TEXTURE_2D = new Texture(Texture.Type.TEXTURE_2D);
    /**
     * Die Fehler 2D Textur.
     */
    public static final Texture ERROR_TEXTURE_2D = new Texture(Texture.Type.TEXTURE_2D);
    /**
     * Alle von Dateien geladenen Texturen, sortiert nach ihrer Resource.
     */
    private static final Map<String, Texture> textures = new HashMap<>();
    
    /**
     * Initialisiert die Standard- und Fehlertextur.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public static boolean init() {
        if(!DEFAULT_TEXTURE_2D.init() || !loadData(DEFAULT_TEXTURE_2D, "textures/default.png")) {
            Logger.INSTANCE.error("DEFAULT_TEXTURE_2D could not be initialized");
            return false;
        }
        DEFAULT_TEXTURE_2D.addRef();
        if(!ERROR_TEXTURE_2D.init() || !loadData(ERROR_TEXTURE_2D, "textures/error.png")) {
            DEFAULT_TEXTURE_2D.delRef();
            Logger.INSTANCE.error("ERROR_TEXTURE_2D could not be initialized");
            return false;
        }
        ERROR_TEXTURE_2D.addRef();
        return true;
    }
    
    /**
     * Zerstoert die Standard- und Fehlertextur.
     */
    public static void destroy() {
        DEFAULT_TEXTURE_2D.delRef();
        ERROR_TEXTURE_2D.delRef();
    }
    
    /**
     * Liefert eine Resource als Textur zurueck und laedt sie dabei vorher,
     * falls es nicht schon passiert ist.
     * @param resource Die Resource
     * @return Die Textur
     */
    public static Texture getTexture(String resource) {
        Texture tex = textures.get(resource);
        if(tex != null) {
            tex.addRef();
            return tex;
        }
        tex = new Texture(Texture.Type.TEXTURE_2D);
        if(!tex.init()) {
            Logger.INSTANCE.error("texture " + resource + " could not be initialized");
            return ERROR_TEXTURE_2D;
        }
        if(!loadData(tex, resource)) {
            Logger.INSTANCE.error("texture " + resource + " could not be loaded");
            return ERROR_TEXTURE_2D;
        }
        textures.put(resource, tex);
        tex.addRef();
        return tex;
    }
    
    /**
     * Erstellt eine Textur aus einem BufferedImage.
     * @param image Das Bild
     * @return Die Textur
     */
    public static Texture createTexture(BufferedImage image) {
        Texture tex = new Texture(Texture.Type.TEXTURE_2D);
        if(!tex.init()) {
            Logger.INSTANCE.error("texture from BufferedImage could not be initialized");
            return ERROR_TEXTURE_2D;
        }
        tex.setData(image.getWidth(), image.getHeight(), image.getColorModel().getNumComponents(), ImageUtil.bufferedImage2floatBuffer(image));
        tex.addRef();
        return tex;
    }
    
    /**
     * Laedt fuer eine Textur ihre Daten aus einer Resource.
     * @param tex Die Textur
     * @param resource Die zu ladenden Daten
     * @return <tt>true</tt>, falls erfolgreich
     */
    private static boolean loadData(Texture tex, String resource) {
        ResHandle handle = GameApp.getResCache().getHandle(new Resource(resource));
        if(handle == null) {
            return false;
        }
        
        if(handle.getExtra() == null || !(handle.getExtra() instanceof ImageExtraData)) {
            return false;
        }
        ImageExtraData extra = (ImageExtraData)handle.getExtra();
        
        FloatBuffer buffer = BufferUtils.createFloatBuffer(handle.getSize() / 4);
        ByteBuffer srcBuffer = ByteBuffer.wrap(handle.getBuffer());
        int max = extra.getWidth() * extra.getHeight() * extra.getChannels();
        for(int i=0; i < max; ++i) {
            buffer.put(srcBuffer.getFloat());
        }
        buffer.position(0);
        handle.unlock();
        
        tex.setData(extra.getWidth(), extra.getHeight(), extra.getChannels(), buffer);
        return true;
    }
}
