/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package gui;

import graphics.geometry.Geometry;
import graphics.geometry.IndexBuffer;
import graphics.geometry.VertexBuffer;
import graphics.shader.Effect;
import gui.components.Component;
import gui.components.Cursor;
import gui.components.Label;
import input.KeyboardHandler;
import input.PointerHandler;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import logging.Logger;
import main.GameApp;
import org.lwjgl.BufferUtils;
import org.lwjgl.input.Mouse;
import org.lwjgl.opengl.GL11;
import view.ScreenElement;

/**
 *
 * @author nico3000
 */
public class Gui extends ScreenElement implements KeyboardHandler, PointerHandler {
    public static final Geometry QUAD = new Geometry();
    public static final Effect GUI_EFFECT = new Effect();
    public static final int EFFECT_TEX_SLOT = 7;
    public static final long MAX_CLICK_MILLIS = 100;
    public static final int MAX_POINTER_BUTTONS = 8;
    
    public static boolean init() {
        IndexBuffer iBuffer = new IndexBuffer();
        VertexBuffer vBuffer = new VertexBuffer();
        if(!GUI_EFFECT.initVF("Shaders/GuiVS.glsl", "Shaders/GuiFS.glsl")) {
            Logger.INSTANCE.error("Initialization of gui effect failed!");
            return false;
        }
        GUI_EFFECT.use();
        GUI_EFFECT.setUniform("g_ComponentTex", EFFECT_TEX_SLOT);
        if(!iBuffer.init() || !vBuffer.init()) {
            Logger.INSTANCE.error("Initialization of gui quad geometry failed!");
            GUI_EFFECT.destroy();
            vBuffer.destroy();
            iBuffer.destroy();
            return false;
        }
        IntBuffer indices = BufferUtils.createIntBuffer(4);
        indices.put(new int[] { 0, 1, 3, 2, });
        indices.position(0);
        iBuffer.setData(indices, IndexBuffer.TRIANGLE_STRIP);
        FloatBuffer vertices = BufferUtils.createFloatBuffer(8);
        vertices.put(new float[] { 0.0f, 0.0f, 
                                   1.0f, 0.0f, 
                                   1.0f, 1.0f, 
                                   0.0f, 1.0f, });
        vertices.position(0);
        vBuffer.setData(vertices);
        vBuffer.addAttribute(VertexBuffer.ATTR_POS, 2);
        if(!QUAD.init(iBuffer, vBuffer)) {
            Logger.INSTANCE.error("Initialization of gui quad failed!");
            GUI_EFFECT.destroy();
            vBuffer.destroy();
            iBuffer.destroy();            
            return false;
        }
        QUAD.addRef();
        return true;
    }
    
    public static void destroy() {
        QUAD.delRef();
        GUI_EFFECT.destroy();
    }

    /**
     * Die Klasse, die fuer das Look and Feel der Gui verantwortlich ist.
     */
    private final ComponentDrawer drawer;
    /**
     * Die Komponenten dieser Gui.
     */
    private List<Component> components = new LinkedList<>();    
    /**
     * Der aktuelle Cursor.
     */
    private final Cursor cursor = new Cursor();
    /**
     * Label fuer Tooltips.
     */
    private final Label tooltip = new Label("Empty Tooltip");
    /**
     * Merkt sich die Komponente unter dem Cursor.
     */
    private Component hovered = null;
    /**
     * Merkt sich die fokussierte Komponentedieser Gui.
     */
    private Component focused = null;
    /**
     * Stati der Mausbuttons.
     */
    private long buttons[] = new long[MAX_POINTER_BUTTONS];
    /**
     * Zeitstempel der letzten Pointerbewegung.
     */
    private long timeOnHovered = 0;

    /**
     * Konstruktor.
     * @param drawer Die Klasse, die fuer das Look&amp;Feel verantwortlich ist
     */
    public Gui(ComponentDrawer drawer) {
        this.drawer = drawer;
        for(int i=0; i < MAX_POINTER_BUTTONS; ++i) {
            this.buttons[i] = -1;
        }
    }
    
    /**
     * Liefert eine Komponente der Gui nach ihrem Namen.
     * @param name Der Name der Komponente
     * @return Die Komponente oder <tt>null</tt>, falls es keine mit dem
     *         angegebenen Namen gibt
     */
    public Component getComponentByName(String name) {
        for(Component component : this.components) {
            if(component.getName().equals(name)) {
                return component;
            }
        }
        return null;
    }
    
    public Cursor getCursor() {
        return this.cursor;
    }

    public ComponentDrawer getDrawer() {
        return drawer;
    }
    
    private void checkHovered(int x, int y) {
        if(this.hovered != null && !this.hovered.isIn(x, y)) {
            this.hovered.onCursorLeft();
            this.hovered = null;
        }
    }
    
    private void leaveHovered() {
        if(this.hovered != null) {
            this.hovered.onCursorLeft();
            this.hovered = null;
        }
    }
    
    private void leaveFocused() {
        if(this.focused != null) {
            this.focused.onFocusLost();
            this.focused = null;
        }
    }
    
    public boolean addComponent(Component comp) {
        if(this.components.contains(comp)) {
            Logger.INSTANCE.warning("Tried to add component to gui twice.");
            return false;
        } else if(!comp.init(this)) {
            Logger.INSTANCE.warning("Component initialization failed.");
            return false;
        } else {
            this.components.add(comp);
            return true;
        }
    }
    
    public boolean removeComponent(Component comp) {
        if(comp == this.cursor) {
            Logger.INSTANCE.warning("Cannot remove own cursor from gui. Use setVisible() instead.");
            return false;
        } else if(!this.components.contains(comp)) {
            Logger.INSTANCE.warning("Tried to remove non-existing component from gui.");
            return false;
        } else {
            comp.destroy();
            this.components.remove(comp);
            return true;
        }
    }

    @Override
    public boolean onKeyDown(int key, char c) {
        return this.focused != null && this.focused.onKeyDown(key, c);
    }

    @Override
    public boolean onKeyUp(int key) {
        return this.focused != null && this.focused.onKeyUp(key);
    }

    @Override
    public boolean isProcessingRepeatEvents() {
        return true;
    }

    @Override
    public void setVisible(boolean visible) {
        super.setVisible(visible);
        if(this.isVisible()) {
            GameApp.getInputHandler().registerKeyboardHandler(this);
            GameApp.getInputHandler().registerPointerHandler(this);
        } else {
            GameApp.getInputHandler().unregisterKeyboardHandler(this);
            GameApp.getInputHandler().unregisterPointerHandler(this);            
        }
        if(this.cursor.isVisible()) {
            Mouse.setGrabbed(visible);
        }
    }

    @Override
    public boolean pointerMoved(int posX, int posY, int dX, int dY) {
        this.tooltip.setPosition(Component.REF_TOP | Component.REF_LEFT, posX + this.cursor.getWidth(), posY - this.cursor.getHeight());
        this.cursor.setPointerPos(posX, posY);
        this.checkHovered(posX, posY);
        
        boolean consumed = false;
        ListIterator<Component> iter = this.components.listIterator(this.components.size());
        while(!consumed && iter.hasPrevious()) {
            Component comp = iter.previous();
            if(comp == this.hovered) {
                comp.onCursorMoved(posX, posY, dX, dY);
                consumed = true;
            } else if(comp.isVisible() && comp.isIn(posX, posY) && comp.onCursorEntered(posX, posY)) {
                this.leaveHovered();
                this.hovered = comp;
                this.timeOnHovered = 0;
                if(comp.getTooltip() != null) {
                    this.tooltip.setSize(8 * comp.getTooltip().length(), 24);
                }
                this.tooltip.setText(comp.getTooltip());
                consumed = true;
            }
        }
        
        if(this.focused != null && this.isDragging()) {
            consumed = false;
            for(int btn=0; !consumed && btn < 8; ++btn) {
                if(this.isButtonDown(btn)) {
                    consumed = this.focused.onDragged(btn, posX, posY, dX, dY);
                }
            }
            return true;
        } else {
            return false;
        }
    }

    @Override
    public boolean onButtonDown(int button) {
        this.buttons[button] = GameApp.getRealTimer().getTimeStamp();
        ListIterator<Component> iter = this.components.listIterator(this.components.size());
        boolean hit = false;
        while(!hit && iter.hasPrevious()) {
            Component comp = iter.previous();
            if(comp.isIn(this.cursor.getPointerX(), this.cursor.getPointerY())) {
                // komponente wurde getroffen
                if(this.focused != comp) {
                    // komponente hat noch nicht den fokus
                    if(comp.onFocusGained()) {
                        // komponente akzeptiert den fokus
                        this.leaveFocused();
                        this.focused = comp;
                        hit = true;
                    } else {
                        // komponente akzeptiert den fokus nicht
                        // weitersuchen
                    }
                } else {
                    // komponente hat bereits fokus
                    hit = true;
                }
            } else {
                // komponente wurde nicht getroffen
                // weitermachen
            }
        }
        if(!hit) {
            // es wurde nichts getroffen
            this.leaveFocused();
            return false;
        } else {
            // es wurde etwas getroffen
            this.focused.onButtonDown(button, this.cursor.getPointerX(), this.cursor.getPointerY());
            return true;
        }
    }

    @Override
    public boolean onButtonUp(int button) {
        boolean clicked = GameApp.getRealTimer().getTimeStamp() - this.buttons[button] < MAX_CLICK_MILLIS;
        this.buttons[button] = -1;
        ListIterator<Component> iter = this.components.listIterator(this.components.size());
        boolean consumed = false;
        while(!consumed && iter.hasPrevious()) {
            Component comp = iter.previous();
            if(comp.isIn(this.cursor.getPointerX(), this.cursor.getPointerY())) {
                // komponente unter cursor
                if(this.focused == comp) {
                    // komponente hat fokus
                    consumed = true;
                } else {
                    // komponente hat fokus nicht
                    // weitermachen
                }
            } else {
                // komponente nicht unter cursor
                // weitermachen
            }
        }
        if(this.focused != null) {
            this.focused.onButtonUp(button, this.cursor.getPointerX(), this.cursor.getPointerY(), clicked);
        }
        return consumed;
    }
    
    private boolean isButtonDown(int button) {
        return this.buttons[button] != -1;
    }
    
    private boolean isDragging() {
        return this.buttons[0] != -1 ||
                this.buttons[1] != -1 ||
                this.buttons[2] != -1 ||
                this.buttons[3] != -1 ||
                this.buttons[4] != -1 ||
                this.buttons[5] != -1 ||
                this.buttons[6] != -1 ||
                this.buttons[7] != -1;
    }

    @Override
    public boolean onWheelMoved(int delta) {
        return false;
    }

    @Override
    public boolean onRestore() {
        this.setVisible(true);
        this.tooltip.setSize(192, 24);
        return this.addComponent(this.cursor) && this.addComponent(this.tooltip);        
    }

    @Override
    public void onDestroy() {
        for(Component comp : this.components) {
            comp.destroy();
        }
        this.components.clear();
    }

    @Override
    public boolean onLostDevice() {
        return true;
    }

    @Override
    public void onRender() {
        GL11.glDisable(GL11.GL_DEPTH_TEST);
        for(Component comp : this.components) {
            if(comp.isVisible()) {
                comp.onRender();
            }
        }
        GL11.glEnable(GL11.GL_DEPTH_TEST);
    }

    @Override
    public void onUpdate(long deltaMillis) {
        this.timeOnHovered += deltaMillis;
        if(this.timeOnHovered < 500 || this.hovered == null || this.tooltip.getText() == null) {
            this.tooltip.setVisible(false);
        } else {
            this.tooltip.setVisible(true);
        }
        Collections.sort(this.components);
        for(Component comp : this.components) {
            if(comp.isVisible()) {
                comp.onUpdate(deltaMillis);
            }
        }
    }
}
