/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package gui.components;

import graphics.shader.UniformBuffer;
import graphics.texture.Texture;
import graphics.texture.TextureLoader;
import gui.Gui;
import gui.GuiAction;
import java.awt.Rectangle;
import java.nio.FloatBuffer;
import java.util.HashSet;
import java.util.Set;
import logging.Logger;
import main.GameApp;
import org.lwjgl.BufferUtils;
import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector4f;
import util.MathUtil;

/**
 * Repraesentiert ein allgemeines Gui Element. Es besteht aus einem
 * positionierten Rechteck gefuellt mit einer Textur.
 * @author nico3000
 */
public abstract class Component implements Comparable<Component> {
    /**
     * Untere Kante der Komponente als Bezugspunkt fuer <tt>setPosition()</tt>.
     */
    public static final int REF_BOTTOM = 1;     // ...0001
    /**
     * Vertikale Mitte der Komponente als Bezugspunkt fuer
     * <tt>setPosition()</tt>.
     */
    public static final int REF_VCENTER = 3;    // ...0011
    /**
     * Obere Kante der Komponente als Bezugspunkt fuer <tt>setPosition()</tt>.
     */
    public static final int REF_TOP = 2;        // ...0010
    /**
     * Linke Kante der Komponente als Bezugspunkt fuer <tt>setPosition()</tt>.
     */
    public static final int REF_LEFT = 4;       // ...0100
    /**
     * Horizontale Mitte der Komponente als Bezugspunkt fuer
     * <tt>setPosition()</tt>.
     */
    public static final int REF_HCENTER = 12;   // ...1100
    /**
     * Rechte Kante der Komponente als Bezugspunkt fuer <tt>setPosition()</tt>.
     */
    public static final int REF_RIGHT = 8;      // ...1000
    /**
     * Maske des vertikalen Bezugspunktes der Komponente fuer
     * <tt>setPosition()</tt>.
     */
    private static final int REF_VER_MASK = 3;  // ...0011
    /**
     * Maske des horizontalen Bezugspunktes der Komponente fuer
     * <tt>setPosition()</tt>.
     */
    private static final int REF_HOR_MASK = 12; // ...1100
   
    /**
     * Temporaere Matrix fuer <tt>updateUniform</tt>.
     */
    private static final Matrix4f TEMP_SCALE = new Matrix4f();
    /**
     * Temporaere Matrix fuer <tt>updateUniform</tt>.
     */
    private static final Matrix4f TEMP_TRANS = new Matrix4f();
    /**
     * Temporaere Matrix fuer <tt>updateUniform</tt>.
     */    
    private static final Matrix4f TEMP_TRANSFORM = new Matrix4f();
    /**
     * Zaehler der Komponenten.
     */
    private static int counter = 0;
    
    /**
     * Der Uniformbuffer der Komponente.
     */
    private final UniformBuffer transform = new UniformBuffer();
    /**
     * Die Daten des Uniformbuffers.
     */
    private final FloatBuffer transformBuffer = BufferUtils.createFloatBuffer(24);
    /**
     * Die Textur der Komponente.
     */
    private Texture texture = TextureLoader.DEFAULT_TEXTURE_2D;
    /**
     * Wurde die Komponente bereits erfolgreich initialisiert?
     */
    private boolean initialized = false;
    /**
     * Die aktuelle Ebene der Komponente.
     */
    private int zOrder = 0;
    /**
     * Gibt an, ob die Komponente sichtbar ist.
     */
    private boolean isVisible = true;
    /**
     * Die Position und Groesse der Komponente als Rechteck in Pixelkoordinaten.
     */
    private Rectangle rect = new Rectangle();
    /**
     * Die Gui dieser Komponente.
     */
    private Gui gui;
    /**
     * Der Name der Komponente.
     */
    private String name;
    /**
     * Tooltip der Komponente.
     */
    private String tooltip = null;
    /**
     * Die Komponente, die diese Komponente enthaelt, sofern vorhanden.
     */
    private Component container = null;
    /**
     * Menge aller Komponenten, die in dieser enthalten sind.
     */
    private final Set<Component> children = new HashSet<>();
    /**
     * Repraesntiert das Clippingrechteck dieser Komponente im projection space.
     */
    private final Vector4f clippingRect = new Vector4f(-1.0f, -1.0f, +1.0f, +1.0f);
    /**
     * Eine mit dieser Komponente verknuepfte Action.
     */
    private GuiAction action = null;
    
    /**
     * Konstruktor.
     */
    public Component() {
        this.texture.addRef();
        this.name = "Component_" + counter++;
    }

    /**
     * Setzt den Namen der Komponente.
     * @param name Der neue Name
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * Liefert den Namen der Komponente.
     * @return Der Name
     */
    public String getName() {
        return this.name;
    }

    /**
     * Setzt einen Tooltip fuer diese Komponente.
     * @param tooltip Der neue Tooltip
     */
    public void setTooltip(String tooltip) {
        this.tooltip = tooltip;
    }

    /**
     * Liefert den Tooltip fuer diese Komponente.
     * @return Der Tooltip
     */
    public String getTooltip() {
        return this.tooltip;
    }

    /**
     * Setzt eine neue Action fuer diese Komponente. Wie die Action aktiviert
     * wird, haengt von der Komponente ab.
     * @param action Die neue Action
     */
    public void setAction(GuiAction action) {
        this.action = action;
    }
    
    /**
     * Feuer die mit dieser ID verknuepfte Action ab.
     * @param optionalData Optionale Zusatzdaten, die von der speziellen Action
     *                     evtl. genutzt werden
     */
    protected void fireAction(Object optionalData) {
        if(this.action != null) {
            this.action.fire(optionalData);
        }
    }
    
    /**
     * Entspricht einem Aufruf von <tt>fireAction(null)<tt>.
     */
    protected void fireAction() {
        this.fireAction(null);
    }
        
    /**
     * Liefert die Gui dieser Komponente.
     * @return Die Gui
     */
    protected Gui getGui() {
        return this.gui;
    }
    
    /**
     * Setzt die lokale Tiefe der Komponente bezueglich des Containers 
     * (initial auf <tt>0</tt>). Die Komponenten werden in aufsteigender
     * Reihenfolge der globalen Tiefe gerendert. Die globale Tiefe ist die Summe
     * der globalen Tiefe des Containers, sofern vorhanden, und der lokalen
     * Tiefe der Komponente.
     * @param zOrder Die neue lokale Tiefe
     */
    public void setLocalZOrder(int zOrder) {
        this.zOrder = zOrder;
    }
    
    /**
     * Liefert die globale Tiefe der Komponente.
     * @return Die globale Tiefe
     */
    public int getGlobalZOrder() {
        return this.container == null ? this.zOrder : this.zOrder + this.container.getGlobalZOrder();
    }
    
    /**
     * Liefert die Sichtbarkeit der Komponente.
     * @return <tt>true</tt>, falls sichtbar
     */
    public boolean isVisible() {
        return (this.container == null || this.container.isVisible()) && this.isVisible;
    }
    
    /**
     * Setzt die Sichtbarkeit der Komponente.
     * @param visible Die neue Sichtbarkeit
     */
    public void setVisible(boolean visible) {
        this.isVisible = visible;
    }

    @Override
    public int compareTo(Component other) {
        return other == null ? -1 : Integer.compare(this.zOrder, other.zOrder);
    }
    
    /**
     * Fuegt dieser Komponente eine Komponente als Kind hinzu und entfernt sie
     * gleichzeitig von ihrem aktuellen Container, sofern vorhanden. Der Aufruf
     * schlaegt fehl, falls die hinzuzufuegende Komponente ein Vorgaenger dieser
     * Komponente oder bereits ein Kind ist oder beide Komponenten nicht Teil
     * derselben Gui sind.
     * @param child Die neue Kindkomponente
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean addChild(Component child) {
        if(this.gui != child.gui) {
            Logger.INSTANCE.warning("Tried to add component (%s) to a component (%s) of another gui.", this.getName(), child.getName());
            return false;
        }
        Component predecessor = this;
        while(predecessor != null) {
            if(predecessor == child) {
                Logger.INSTANCE.error("Tried to add a predecessor (%s) of a node (%s) as its child.", child.getName(), this.getName());
                return false;
            }
            predecessor = child.container;
        }
        if(!this.children.add(child)) {
            Logger.INSTANCE.error("Tried to add a component (%s) as a child of another component (%s) twice.", child.getName(), this.getName());
            return false;
        } else {
            if(child.container != null) {
                child.container.children.remove(child);
            }
            child.container = this;
            child.dimensionChanged(false);
            return true;
        }
    }
    
    /**
     * Entspricht einem Aufruf von <tt>setSize(width, height)</tt> und
     * anschliessendem <tt>setPosition(x, y)</tt>.
     * @param x Der Abstand vom linken (positiv) oder rechten (negativ) 
     *          Containerrand
     * @param y Der Abstand vom unteren (positiv) oder oberen (negativ)
     *          Containerrand
     * @param width Die Breite
     * @param height Die Hoehe
     */
    public void setDimension(int x, int y, int width, int height) {
        this.setSize(width, height);
        this.setPosition(x, y);
    }
    
    /**
     * Setzt die Position des Mittelpunkts der Komponente relativ zum Container.
     * @param percentageX Der relative Abstand vom linken (positiv) bzw. rechten
     *                    (negativ) Fenster- bzw. Containerrand
     * @param percentageY Der relative Abstand vom oberen (positiv) bzw. unteren
     *                    (negativ) Fenster- bzw. Containerrand
     */
    public void setPosition(float percentageX, float percentageY) {
        int x = (int)(percentageX * (float)this.getContainerWidth());
        int y = (int)(percentageY * (float)this.getContainerHeight());
        this.setPosition(REF_HCENTER | REF_VCENTER, x, y);
    }
    
    /**
     * Setzt die Positionen relativ zum Fensterrand. Positive Werte werden als
     * Abstand vom linken bzw. unteren Rand interpretiert, waehrend negative
     * Werte als Abstand vom rechten, bzw. oberen Rand interpretiert werden.
     * Nach einem Aufruf von <tt>setSize()</tt> muessen diese Werte evtl. neu
     * gesetzt werden. Der Nullpunkt ist dabei der linke untere Rand des
     * Fensters bzw. Containers.
     * @param x Der Abstand vom linken (positiv) oder rechten (negativ) Fenster-
     *          bzw. Containerrand
     * @param y Der Abstand vom unteren (positiv) oder oberen (negativ) Fenster-
     *          bzw. Containerrand
     */
    public void setPosition(int x, int y) {
        int ref = (x < 0 ? REF_RIGHT : REF_LEFT) | (y < 0 ? REF_TOP : REF_BOTTOM);
        this.setPosition(ref, x, y);
    }
    
    /**
     * Setzt die Position der Komponente anhand eines Referenzpunktes. Der
     * Nullpunkt ist dabei der linke untere Rand des Fensters bzw. Containers.
     * @param reference Der Referenzpunkt der Komponente; eine Kombination aus
     *                  <tt>REF_BOTTOM</tt>, <tt>REF_VCENTER</tt>,
     *                  <tt>REF_TOP</tt>, <tt>REF_LEFT</tt>,
     *                  <tt>REF_HCENTER</tt> und <tt>REF_RIGHT</tt>
     * @param x Die neue x-Koordinate des Referenzpunktes
     * @param y Die neue y-Koordinate des Referenzpunktes 
     */
    public void setPosition(int reference, int x, int y) { 
        if(x < 0) {
            x += this.getContainerWidth();
        }
        if(y < 0) {
            y += this.getContainerHeight();
        }        
        switch(reference & REF_VER_MASK) {
            case REF_TOP:
                y -= this.rect.height;
                break;
            case REF_VCENTER:
                y -= this.rect.height / 2;
                break;
        }
        switch(reference & REF_HOR_MASK) {
            case REF_RIGHT:
                x -= this.rect.width;
                break;
            case REF_HCENTER:
                x -= this.rect.width / 2;
                break;
        }
        this.setAbsolutePosition(x, y);
    }
    
    /**
     * Setzt die Position der linken unteren Ecke der Komponente. Der Nullpunkt
     * ist dabei der linke untere Rand des Fensters oder, sofern vorhanden, des
     * Containers.
     * @param x Die neue x-Koordinate
     * @param y Die neue y-Koordinate
     */
    public void setAbsolutePosition(int x, int y) {
        this.rect.x = x;
        this.rect.y = y;
        this.dimensionChanged(false);
    }
    
    /**
     * Setzt die Groesse der Komponente in Pixel.
     * @param width Die Breite
     * @param height Die Hoehe
     */
    public void setSize(int width, int height) {
        this.rect.width = width;
        this.rect.height = height;
        this.dimensionChanged(true);
    }
    
    /**
     * Setzt die eigene Groesse prozentual zur Fenster- / Containergroesse.
     * @param percentageWidth Die Breite
     * @param percentageHeight Die Hoehe
     */
    public void setSize(float percentageWidth, float percentageHeight) {
        int width = this.getContainerWidth();
        int height = this.getContainerHeight();
        this.setSize((int)(percentageWidth * (float)width), (int)(percentageHeight * (float)height));
    }
    
    /**
     * Liefert die Breite in Pixel der Komponte.
     * @return Die Breite
     */
    public int getWidth() {
        return this.rect.width;
    }
    
    /**
     * Liefert die Hoehe in Pixel der Komponte.
     * @return Die Hoehe
     */
    public int getHeight() {
        return this.rect.height;
    }
    
    /**
     * Existiert eine Elternkomponente wird dessen Breite zurueckgegeben,
     * ansonsten Breite des Fensters.
     * @return Die Breite des Containers
     */
    private int getContainerWidth() {
        return this.container == null ? GameApp.getGraphicsLayer().getWidth() : this.container.getWidth();
    }
    
    /**
     * Existiert eine Elternkomponente wird dessen Hoehe zurueckgegeben,
     * ansonsten Hoehe des Fensters.
     * @return Die Hoehe des Containers
     */
    private int getContainerHeight() {
        return this.container == null ? GameApp.getGraphicsLayer().getHeight(): this.container.getHeight();
    }
    
    /**
     * Liefert die globale x-Koordinate der linken unteren Ecke dieser
     * Komponente. Der Nullpunkt ist dabei die linke untere Ecke des Fensters.
     * @return Die x-Koordinate
     */
    public int getGlobalX() {
        if(this.container == null) {
            return this.rect.x;
        } else {
            return this.rect.x + this.container.getGlobalX();
        }
    }
    
    /**
     * Liefert die globale y-Koordinate der linken unteren Ecke dieser
     * Komponente. Der Nullpunkt ist dabei die linke untere Ecke des Fensters.
     * @return Die y-Koordinate
     */
    public int getGlobalY() {
        if(this.container == null) {
            return this.rect.y;
        } else {
            return this.rect.y + this.container.getGlobalY();
        }
    }
    
    /**
     * Wird aufgerufen, wenn der Cursor sich auf die Komponente bewegt hat.
     * @param x Die x-Koordinate des Cursors
     * @param y Die y-Koordinate des Cursors
     * @return <tt>true<tt>, falls die Komponente die Aktion akzeptiert
     */
    public abstract boolean onCursorEntered(int x, int y);
    
    /**
     * Wird aufgerufen, wenn der Cursor von der Komponente entfernt wird.
     */  
    public abstract void onCursorLeft();
    
    /**
     * Wird aufgerufen, wenn der Cursor sich auf der Komponente bewegt hat.
     * @param x Die neue absolute x Position des Cursors
     * @param y Die neue absolute y Position des Cursors
     * @param dx Die Bewegung des Cursors in x Richtung
     * @param dy Die Bewegung des Cursors in y Richtung
     */
    public abstract void onCursorMoved(int posX, int posY, int dx, int dy);
    
    /**
     * Wird aufgerufen, wenn die Komponente den Fokus bekommen soll.
     * @return <tt>true<tt>, falls die Komponente die Aktion akzeptiert
     */    
    public abstract boolean onFocusGained();
    
    /**
     * Wird aufgerufen, wenn die Komponente den Fokus verliert.
     */  
    public abstract void onFocusLost();
    
    /**
     * Wird aufgerufen, wenn eine Taste des Cursors auf der Komponente gedrueckt
     * wird.
     * @param button Die Taste
     * @param x Die x-Koordinate des Cursors
     * @param y Die y-Koordinate des Cursors
     * @return <tt>true<tt>, falls die Komponente die Aktion akzeptiert
     */
    public abstract boolean onButtonDown(int button, int x, int y);
    
    /**
     * Wird aufgerufen, wenn eine Taste des Cursors auf der Komponente
     * losgelassen wird.
     * @param button Die Taste
     * @param x Die x-Koordinate des Cursors
     * @param y Die y-Koordinate des Cursors
     * @param clicked <tt>true</tt>, falls die Zeit, in der die Taste unten war,
     *                einen bestimmten Wert nicht ueberschreitet und somit als
     *                Klick interpretiert werden kann
     * @return <tt>true<tt>, falls die Komponente die Aktion akzeptiert
     */
    public abstract boolean onButtonUp(int button, int x, int y, boolean clicked);
    
    /**
     * Wird aufgerufen, wenn der Cursor die Komponente mit gedrueckter Taste
     * zieht.
     * @param button Der Button der Maus
     * @param x Die neue absolute x Position des Cursors
     * @param y Die neue absolute y Position des Cursors
     * @param dx Die Bewegung des Cursors in x Richtung
     * @param dy Die Bewegung des Cursors in y Richtung
     * @return <tt>true<tt>, falls die Komponente die Aktion akzeptiert
     */
    public abstract boolean onDragged(int button, int x, int y, int dx, int dy);

    /**
     * Wird waehrend jedem Frame aufgerufen.
     * @param deltaMillis Die Dauer des letzten Frames in Milliskunden
     */
    public abstract void onUpdate(long deltaMillis);
    
    /**
     * Wird aufgerufen, wenn die Komponente den Fokus hat und eine Taste der
     * Tastatur gedrueckt wurde.
     * @param key Die Taste
     * @param c Der zugehoerige Character, falls vorhanden; sonst <tt>0</tt>
     * @return <tt>true</tt>, falls die Komponente die Aktion akzeptiert
     */
    public abstract boolean onKeyDown(int key, char c);
    
    /**
     * Wird aufgerufen, wenn die Komponente den Fokus hat und eine Taste der
     * Tastatur losgelassen wurde.
     * @param key Die Taste
     * @return <tt>true</tt>, falls die Komponente die Aktion akzeptiert
     */
    public abstract boolean onKeyUp(int key);
    
    /**
     * Prueft, ob ein Punkt innerhalb der Komponente liegt.
     * @param x Die x-Koordinate
     * @param y Die y-Koordinate
     * @return <tt>true</tt>, falls innerhalb
     */
    public boolean isIn(int x, int y) {
        if(this.container == null) {
            return this.rect.contains(x, y);
        } else {
            return this.container.isIn(x, y) && this.rect.contains(x - this.container.getGlobalX(), y - this.container.getGlobalY());
        }
    }
    
    /**
     * Zerstoert die Komponente.
     */
    public void destroy() {
        this.transform.destroy();
        if(this.texture != null) {
            this.texture.addRef();
        }
    }

    /**
     * Setzt die Textur der Komponente.
     * @param texture Die neue Textur
     */
    public void setTexture(Texture texture) {
        if(this.texture != null) {
            this.texture.delRef();
        }
        this.texture = texture;
        if(this.texture != null) {
            this.texture.addRef();
        }
    }
    
    /**
     * Wird aufgerufen, wenn die Komponente ihre Textur aktualisieren soll.
     */
    protected abstract void repaint();
    
    /**
     * Wird aufgerufen, wenn sich die Groesse oder Position der Komponent
     * geaendert hat.
     */
    private void dimensionChanged(boolean sizeChanged) {
        if(this.initialized) {
            float transX = -1.0f + 2.0f * (float)this.getGlobalX() / GameApp.getGraphicsLayer().getWidth();
            float transY = -1.0f + 2.0f * (float)this.getGlobalY() / GameApp.getGraphicsLayer().getHeight();
            float scaleX = 2.0f * (float)this.getWidth() / GameApp.getGraphicsLayer().getWidth();
            float scaleY = 2.0f * (float)this.getHeight() / GameApp.getGraphicsLayer().getHeight();
            this.clippingRect.set(transX, transY, transX + scaleX, transY + scaleY);
            MathUtil.scaling(Component.TEMP_SCALE, scaleX, scaleY, 1.0f);
            MathUtil.translation(Component.TEMP_TRANS, transX, transY, 0.0f);
            MathUtil.mul(Component.TEMP_TRANSFORM, Component.TEMP_TRANS, Component.TEMP_SCALE);
            Component.TEMP_TRANSFORM.store(this.transformBuffer);
            if(this.container != null) {
                this.container.clippingRect.store(this.transformBuffer);
            } else {
                this.transformBuffer.put(new float[] { -1.0f, -1.0f, +1.0f, +1.0f, });
            }
            this.transformBuffer.position(0);
            this.transform.setData(this.transformBuffer);
            for(Component child : this.children) {
                child.dimensionChanged(false);
            }
            if(sizeChanged) {
                this.repaint();
            }
        }
    }

    /**
     * Initialisiert die Komponente.
     * @param gui Die Gui
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init(Gui gui) {
        this.gui = gui;
        this.transform.setBindingIndex(UniformBuffer.BINDING_MODEL);
        if(!this.transform.init()) {
            Logger.INSTANCE.error("Initialization of gui uniform buffer failed!");
            return false;
        }
        this.initialized = true;
        return true;
    }
    
    /**
     * Zeichnet die Komponente.
     */
    public void onRender() {
        Gui.GUI_EFFECT.use();
        this.transform.bind();
        this.texture.connectToSlot(Gui.EFFECT_TEX_SLOT);
        this.texture.bind();
        Gui.QUAD.bind();
        Gui.QUAD.render();
    }
}
