/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package input;

import java.util.Iterator;
import java.util.LinkedList;
import logging.Logger;
import org.lwjgl.LWJGLException;
import org.lwjgl.input.Keyboard;
import org.lwjgl.input.Mouse;

/**
 * Liest Events von Tastatur und Maus und leitet sie an die registrierten
 * Handlerklassen weiter.
 * @author nico3000
 */
public class InputHandler {
    /**
     * Rueckgabewert der Methode <tt>getKey()</tt> bei ungueltigem Parameter.
     */
    public static final int NO_KEY = Keyboard.KEY_NONE;
    
    /**
     * Rueckgabewert der Methode <tt>getButton()</tt> bei ungueltigem Parameter.
     */
    public static final int NO_BUTTON = -1;
    
    /**
     * Die registrierten KeyboardHandler.
     */
    private final LinkedList<KeyboardHandler> keyboardHandlers = new LinkedList<>();
    
    /**
     * Die registrierten PointerHandler.
     */
    private final LinkedList<PointerHandler> pointerHandlers = new LinkedList<>();
    
    /**
     * Initialisiert die Maus und Tastatur. Ob8: Vorher muss
     * <tt>Display.create()</tt> aufgerufen worden sein.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init() {
        try {
            Keyboard.create();
            Keyboard.enableRepeatEvents(true);
            Mouse.create();
            return true;
        } catch (LWJGLException ex) {
            Logger.INSTANCE.error("keyboard and/or mouse could not be created: " + ex.getMessage());
            return false;
        }
    }
    
    /**
     * Gibt Maus und Tastatur wieder frei.
     */
    public void destroy() {
        Mouse.destroy();
        Keyboard.destroy();
    }
    
    /**
     * Verarbeitet alle neuen Maus- und Tastaturevents und gibt sie an die
     * entsrpechenden Handler weiter.
     */
    public void processInput() {
        // process keyboard
        while(Keyboard.next()) {
            boolean processed = false;
            boolean down = Keyboard.getEventKeyState();
            Iterator<KeyboardHandler> iter = this.keyboardHandlers.iterator();
            while(!processed && iter.hasNext()) {
                KeyboardHandler handler = iter.next();
                if(down) {
                    if(!Keyboard.isRepeatEvent() || handler.isProcessingRepeatEvents()) {
                        processed = handler.onKeyDown(Keyboard.getEventKey(), Keyboard.getEventCharacter());
                    }
                } else {
                    processed = handler.onKeyUp(Keyboard.getEventKey());
                }
            }
        }
        
        // process mouse
        while(Mouse.next()) {
            boolean processed = false;
            int button = Mouse.getEventButton();
            boolean down = Mouse.getEventButtonState();
            int wheel = Mouse.getEventDWheel();
            int posX = Mouse.getEventX();
            int posY = Mouse.getEventY();
            int dX = Mouse.getEventDX();
            int dY = Mouse.getEventDY();
            Iterator<PointerHandler> iter = this.pointerHandlers.iterator();
            while(!processed && iter.hasNext()) {
                if(button == -1 && (dX != 0 || dY != 0)) {
                    processed = iter.next().pointerMoved(posX, posY, dX, dY);
                } else if(wheel != 0) {
                    processed = iter.next().onWheelMoved(wheel);
                } else if(down) {
                    processed = iter.next().onButtonDown(button);
                } else {
                    processed = iter.next().onButtonUp(button);             
                }
            }
        }
    }
    
    /**
     * Registriert einen PointerHandler.
     * @param handler Der Handler
     * @return <tt>true</tt>, falls der Handler noch nicht registriert war;
     * <tt>false</tt> sonst.
     */
    public boolean registerPointerHandler(PointerHandler handler) {
        if(this.pointerHandlers.contains(handler)) {
            Logger.INSTANCE.warning("Tried to register PointerHandler twice.");
            return false;
        } else {
            this.pointerHandlers.addFirst(handler);
            return true;
        }
    }
    
    /**
     * Registriert einen KeyboardHandler.
     * @param handler Der Handler
     * @return <tt>true</tt>, falls der Handler noch nicht registriert war;
     * <tt>false</tt> sonst.
     */
    public boolean registerKeyboardHandler(KeyboardHandler handler) {
        if(this.keyboardHandlers.contains(handler)) {
            Logger.INSTANCE.warning("Tried to register KeyboardHandler twice.");
            return false;
        } else {
            this.keyboardHandlers.addFirst(handler);
            return true;
        }
    }
    
    /**
     * Entfernt einen PointerHandler.
     * @param handler Der Handler
     * @return <tt>true</tt>, falls der Handler registriert war; <tt>false</tt>
     * sonst.
     */
    public boolean unregisterPointerHandler(PointerHandler handler) {
        if(!this.pointerHandlers.remove(handler)) {
            Logger.INSTANCE.warning("Tried to unregister non-existent PointerHandler.");
            return false;
        } else {
            return true;
        }            
    }
    
    /**
     * Entfernt einen KeyboardHandler.
     * @param handler Der Handler
     * @return <tt>true</tt>, falls der Handler registriert war; <tt>false</tt>
     * sonst.
     */
    public boolean unregisterKeyboardHandler(KeyboardHandler handler) {
        if(!this.keyboardHandlers.remove(handler)) {
            Logger.INSTANCE.warning("Tried to unregister non-existent KeyboardHandler.");
            return false;
        } else {
            return true;
        }    
    }
    
    /**
     * Liefert den Keycode zu einer Taste auf der Tastatur anhand ihres Namens.
     * Ist der Name ungueltig wird <tt>NO_KEY</tt> zurueckgegeben.
     * @param name Der Name des Keys
     * @return Der zugehoerige Keycode
     */
    public int getKey(String name) {
        return Keyboard.getKeyIndex(name);
    }
    
    /**
     * Liefert die Nummer des Buttons der Maus zu seinem Namen. Ist der Name
     * ungueltig wird <tt>NO_BUTTON</tt> zurueckgegeben.
     * @param name Der Name des Buttons
     * @return Die Nummer des Buttons
     */
    public int getButton(String name) {
        return Mouse.getButtonIndex(name);
    }
}
