/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package mesh;

import graphics.shader.Effect;
import graphics.shader.UniformBuffer;
import graphics.texture.SlotAdmin;
import graphics.texture.Texture;
import graphics.texture.TextureLoader;

import java.nio.FloatBuffer;

import logging.Logger;

import org.lwjgl.BufferUtils;
import org.lwjgl.util.vector.Vector4f;

/**
 * Ein Material eines Meshes.
 * @author nico3000
 */
public class Material {
    public static final int AMBIENT_TEX_SLOT = SlotAdmin.getInstance().getFreeSlot();
    public static final int DIFFUSE_TEX_SLOT = SlotAdmin.getInstance().getFreeSlot();
    public static final int SPECULAR_TEX_SLOT = SlotAdmin.getInstance().getFreeSlot();
    public static final int ALPHA_TEX_SLOT = SlotAdmin.getInstance().getFreeSlot();
    /**
     * Der Name des Materials.
     */
    private final String name;
    /**
     * Reference Counter.
     */
    private int ref = 0;
    /**
     * UniformBuffer der Daten im Shader.
     */
    private final UniformBuffer buffer = new UniformBuffer();
    /**
     * Die ambiente Farbe.
     */
    private final Vector4f ka = new Vector4f();
    /**
     * Die diffuse Farbe.
     */
    private final Vector4f kd = new Vector4f();
    /**
     * Die spekulare Farbe.
     */
    private final Vector4f ks = new Vector4f();
    /**
     * Textur fuer den Alphakanal.
     */
    private Texture mapAlpha = null;
    /**
     * Textur fuer den ambienten Anteil.
     */
    private Texture mapAmbient = null;
    /**
     * Textur fuer den diffusen Anteil.
     */
    private Texture mapDiffuse = null;
    /**
     * Textur fuer den spekularen Anteil.
     */
    private Texture mapSpecular = null;
    
    private String resourceAlpha = null;
    private String resourceAmbient = null;
    private String resourceDiffuse = null;
    private String resourceSpecular = null;
    private int size = 0;

    /**
     * Konstruktor.
     * @param name Der Name des Materials
     */
    public Material(String name) {
        this.name = name;
    }

    /**
     * Liefert den Namen des Materials.
     * @return Der Name
     */
    public String getName() {
        return name;
    }
    
    /**
     * Initialisiert den Buffer des Materials.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init() {
        boolean success = this.buffer.init();
        if(success) {
            this.buffer.setBindingIndex(UniformBuffer.BINDING_MATERIAL);
        }
        if(this.resourceAlpha != null && (this.mapAlpha = TextureLoader.getTexture(this.resourceAlpha)) == null) {
            Logger.INSTANCE.error("could not load material alpha texture: " + this.resourceAlpha);
            return false;
        }
        if(this.resourceAmbient != null && (this.mapAmbient = TextureLoader.getTexture(this.resourceAmbient)) == null) {
            Logger.INSTANCE.error("could not load material ambient texture: " + this.resourceAmbient);
            return false;
        }
        if(this.resourceDiffuse != null && (this.mapDiffuse = TextureLoader.getTexture(this.resourceDiffuse)) == null) {
            Logger.INSTANCE.error("could not load material diffuse texture: " + this.resourceDiffuse);
            return false;
        }
        if(this.resourceSpecular != null && (this.mapSpecular = TextureLoader.getTexture(this.resourceSpecular)) == null) {
            Logger.INSTANCE.error("could not load material specular texture: " + this.resourceSpecular);
            return false;
        }
        this.size = 16 * 4 + (this.mapAlpha != null ? this.mapAlpha.getSize() : 0)
                           + (this.mapAmbient != null ? this.mapAmbient.getSize() : 0)
                           + (this.mapDiffuse != null ? this.mapDiffuse.getSize() : 0)
                           + (this.mapSpecular != null ? this.mapSpecular.getSize() : 0);
        return success;
    }
    
    /**
     * Bindet das Material an den Uniformblock <tt>MaterialBlock</tt>.
     */
    public void bind() {
        if(this.mapAmbient != null) {
            this.mapAmbient.connectToSlot(AMBIENT_TEX_SLOT);
        }
        if(this.mapDiffuse != null) {
            this.mapDiffuse.connectToSlot(DIFFUSE_TEX_SLOT);
        }
        if(this.mapSpecular != null) {
            this.mapSpecular.connectToSlot(SPECULAR_TEX_SLOT);
        }
        if(this.mapAlpha != null) {
            this.mapAlpha.connectToSlot(ALPHA_TEX_SLOT);
        }
        
        this.buffer.bind();
    }
    
    /**
     * Updated den Buffer des Materials. Muss aufgerufen werden, wenn
     * Aenderungen an den Farben vorgenommen wurden.
     */
    public void updateBuffer() {
        FloatBuffer data = BufferUtils.createFloatBuffer(16);
        this.ka.store(data);
        this.kd.store(data);
        this.ks.store(data);
        data.put(this.mapAmbient == null ? 0.0f : 1.0f);
        data.put(this.mapDiffuse == null ? 0.0f : 1.0f);
        data.put(this.mapSpecular == null ? 0.0f : 1.0f);
        data.put(this.mapAlpha == null ? 0.0f : 1.0f);
        data.position(0);
        this.buffer.setData(data);
    }
    
    /**
     * Bindet die Variablen eines Effekts an die entsprechenden Slots fuer die
     * Texturen.
     * @param effect Der Effekt
     */
    public static void modifyEffectToUseMaterial(Effect effect) {
        effect.use();
        effect.setUniform("g_AmbientTex", Material.AMBIENT_TEX_SLOT);
        effect.setUniform("g_DiffuseTex", Material.DIFFUSE_TEX_SLOT);
        effect.setUniform("g_SpecularTex", Material.SPECULAR_TEX_SLOT);
        effect.setUniform("g_AlphaTex", Material.ALPHA_TEX_SLOT);
    }
    
    /**
     * Setzt eine Resource fuer die ambiente Textur.
     * @param resource Der Name der Resource
     */
    public void setAmbientTextue(String resource) {
        this.resourceAmbient = resource;
    }
    
    /**
     * Setzt eine Resource fuer die diffuse Textur.
     * @param resource Der Name der Resource
     */
    public void setDiffuseTextue(String resource) {
        this.resourceDiffuse = resource;
    }
    
    /**
     * Setzt eine Resource fuer die spekulare Textur.
     * @param resource Der Name der Resource
     */
    public void setSpecularTextue(String resource) {
        this.resourceSpecular = resource;
    }
    
    /**
     * Setzt eine Resource fuer die Alphakanaltextur.
     * @param resource Der Name der Resource
     */
    public void setAlphaTextue(String resource) {
        this.resourceAlpha = resource;
    }

    /**
     * Liefert die ambiente Farbe.
     * @return Die ambiente Farbe
     */
    public Vector4f getKa() {
        return ka;
    }

    /**
     * Liefert die diffuse Farbe.
     * @return Die diffuse Farbe
     */
    public Vector4f getKd() {
        return kd;
    }

    /**
     * Liefert die spekulare Farbe.
     * @return Die spekulare Farbe
     */
    public Vector4f getKs() {
        return ks;
    }
    
    /**
     * Erhoeht den Reference Counter um 1.
     */
    public void addRef() {
        ++this.ref;
    }

    public int getSize() {
        return size;
    }
    
    /**
     * Verringert den Reference Counter um 1. Ist er danach kleiner oder gleich
     * 0, wird das Material zerstoert.
     */
    public void delRef() {
        if(--this.ref <= 0) {
            this.buffer.destroy();
            if(this.mapAlpha != null) {
                this.mapAlpha.delRef();
            }
            if(this.mapAmbient != null) {
                this.mapAmbient.delRef();
            }            
            if(this.mapDiffuse != null) {
                this.mapDiffuse.delRef();
            }
            if(this.mapSpecular != null) {
                this.mapSpecular.delRef();
            }            
        }
    }
    
    @Override
    public String toString() {
    	// TODO Auto-generated method stub
    	return this.name + " Kd :" + kd.toString();
    }
}
