/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package mesh;

import graphics.geometry.Geometry;
import graphics.geometry.IndexBuffer;
import graphics.geometry.VertexBuffer;

import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import logging.Logger;
import mesh.obj.MtlData;
import mesh.obj.ObjData;

import org.lwjgl.BufferUtils;

import util.GeometryFactory;

/**
 * Ein mesh ist ein Objekt im dreidimensionalen Raum mit Materialeigenschaften.
 * @author nico3000
 */
public class Mesh {    
    /**
     * Der Name der Resource.
     */
    private final String resource;
    /**
     * Die Intervalle des Indexbuffers gehoeren zu bestimmten Materialien.
     */
    private Map<Material, IndexBufferInterval> intervals = new HashMap<>();
    /**
     * Die Geometriedaten des Meshes.
     */
    private Geometry geometry;
    
    private Map<String, MtlData> materialLibraries = new HashMap<>();
    
    private MtlData currentMaterialLibrary = null;
    
    /**
     * Reference counter.
     */
    private int ref = 0;
    /**
     * Solange das Mesh nicht geladen ist, der Zeitstempel des Erstellens, sonst
     * die Ladezeit in Millisekunden.
     */
    private long loadingTime = System.currentTimeMillis();
    /**
     * Die Anzahl der Vertices.
     */
    private int vertexCount = 0;
    /**
     * Die Anzahl der Dreiecke.
     */
    private int triangleCount = 0;
    /**
     * Groesse des Meshes im VRAM.
     */
    private int size = 0;
    
    public Mesh(Mesh otherMesh) {
    	this.resource = otherMesh.resource;
    	this.intervals = otherMesh.intervals;
    	this.geometry = otherMesh.geometry;
    	this.materialLibraries = otherMesh.materialLibraries;
    	this.currentMaterialLibrary = otherMesh.currentMaterialLibrary;
    	this.geometry.addRef();
    }
    
    /**
     * Konstruktor eines Meshs aus einer Resource.
     * @param resource Die Name der Resource
     */
    Mesh(String resource) {
        this.resource = resource;
        this.geometry = new Geometry();
        this.geometry.addRef();
    }

    /**
     * Konstruktor eines Meshs aus einer Geometrie.
     * @param geometry Die Geometrie
     */
    Mesh(Geometry geometry) {
        this.resource = "copy";
        this.geometry = geometry;
        this.geometry.addRef();
    }
    
    /**
     * Zeichnet das Mesh. Dazu wird zunaechst die Gemetrie gebunden und danach
     * werden die einzelnen Teile mit ihrem jeweiligem Materiali nacheinander
     * gezeichnet.
     */
    public void draw() {
        this.geometry.bind();
        if(this.intervals.isEmpty()) {
            this.geometry.render();
        } else {
            for(Material material : this.intervals.keySet()) {
                if(material != null) {
                    material.bind();
                }
                IndexBufferInterval interval = this.intervals.get(material);
                this.geometry.draw(interval.getStart(), interval.getCount());
            }
        }            
    }
    
    /**
     * Interpretiert die Resource als obj-Datei und laedt die Daten der
     * Geometrie. Im Anschluss sollte <tt>initGeometry()</tt> ausgefuehrt
     * werden.
     * @param mirrored Soll das Mesh an der x-Achse gespiegelt werden?
     * @return <tt>true</tt>, wenn erfolgreich
     */
    public boolean initMesh(boolean mirrored) {
        ObjData data = new ObjData(this.resource);
        
        boolean success = false;
        if(!data.init()) {
            return false;
        } else {
        	this.geometry = GeometryFactory.buildFromObjData(data, mirrored);
        	if (this.geometry != null) {
        		success = true;
        	}
        }
        
        this.materialLibraries = data.getMaterialLibraries();
        
        List<Integer> indices = new LinkedList<>();
        Map<ObjData.ObjIndex, Integer> indexMappings = new HashMap<>();
        int intervalStart = 0;
        for(Material material : data.getMaterials()) {
            for(ObjData.Face face : data.getFacesByMaterial(material)) {
                this.triangleCount += face.getIndexCount() - 2;
                for(int i=0; i < face.getIndexCount(); ++i) {
                    ObjData.ObjIndex objIndex = face.getIndex(mirrored ? face.getIndexCount() - 1 - i : i);
                    Integer index = indexMappings.get(objIndex);
                    indices.add(index);
                }
                indices.add(IndexBuffer.RESTART);
            }
            this.intervals.put(material, new IndexBufferInterval(intervalStart, indices.size() - intervalStart));
            if(material != null) {
            	if(material.init()) {
                    material.updateBuffer();
                } else {
                    Logger.INSTANCE.error("material " + material.getName() + " could not be initialized");
                    this.destroy();
                    return false;
                }
                material.addRef();
            }
            intervalStart = indices.size();
        }
        data.destroy();
        return true;
    }
    
    public boolean setMtlLib (String library) {
    	// TODO Ausgaben, IndexBuffer nicht durchlaufen, wenn Material schon richtig
    	boolean success = false;
    	
    	if (library == null || library.equals("")) {
    		library = "default";
    	}

    	if (currentMaterialLibrary != null && currentMaterialLibrary.getResource().equals(library)) {
    		return true;
    	}
    	    	
    	MtlData materialLibrary = this.materialLibraries.get(library);
    	String libraryName = library.substring(7, library.length() - 4);
    	
    	if (materialLibrary == null && library.contains("default")) {
	    	for (MtlData matlibrary : this.materialLibraries.values()) {	    		
	    		String resourceName = this.resource.substring(7, this.resource.length() - 4);
	    		if (libraryName.equals(resourceName)) {
	    			materialLibrary = matlibrary;
	    		}
	    	}
    	}
    	
    	if (materialLibrary != null) {
    		this.currentMaterialLibrary = materialLibrary;
    		success = true;
	    	Map<Material, IndexBufferInterval> newIntervals = new HashMap<>();
	    	for (Material material : this.intervals.keySet()) {
	    		if (material != null) {
		    		Material newMaterial = this.currentMaterialLibrary.getMaterial(material.getName());
		    		if (newMaterial != null) {
		    			newMaterial.init();
		    			newMaterial.updateBuffer();
		    			newMaterial.addRef();
		    			newIntervals.put(newMaterial, this.intervals.get(material));
		    		} else {
		    			newIntervals.put(material, this.intervals.get(material));
		    		}
	    		} else {
	    			newIntervals.put(material, this.intervals.get(material));
	    		}
	    	}
	    	
	    	if (success == true) {
	    		this.intervals = newIntervals;
	    	}
    	}
    	
    	return success;
    }
    
    /**
     * Erhoeht den Reference Counter um 1.
     */
    public void addRef() {
        ++this.ref;
    }
    
    /**
     * Verringert den Reference Counter um 1 und loescht die Resource, falls er
     * kleiner oder gleich 0 wird.
     */
    public void delRef() {
        if(--this.ref <= 0) {
            this.destroy();
        }
    }
    
    /**
     * Liefert zurueck, ob das Mesh noch geladen ist und angezeigt werden kann.
     * @return <tt>true</tt>, falls alles in Ordnung ist
     */
    public boolean isOk() {
        return this.ref > 0;
    }
    
    /**
     * Zerstoert diese Resource.
     */
    private void destroy() {
        this.geometry.delRef();
        for(Material material : this.intervals.keySet()) {
            if(material != null) {
                material.delRef();
            }
        }
    }
    
    /**
     * Liefert einen String mit Statistiken des Meshes.
     * @return Statistiken
     */
    public String getStats() {
        return String.format("%s stats: %d vertices; %d triangles; %d materials; %.2fs loading time",
                             this.resource, this.vertexCount, this.triangleCount,
                             this.intervals.size(), 1e-3f * (float)(this.loadingTime));
    }
}
