/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package mesh.obj;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;
import logging.Logger;
import main.GameApp;
import mesh.Material;
import org.lwjgl.util.vector.Vector4f;
import resource.ResHandle;
import resource.Resource;

/**
 * Haelt die Daten einer Wavefront mtl-Datei.
 * @author nico3000
 */
public class MtlData {
    private final String resource;
    private Material currentMaterial = null;
    private int currentLine;
    private final Map<String, Material> materials = new HashMap<>();
    
    /**
     * Konstruktor.
     * @param resource Der Name der Resource
     */
    public MtlData(String resource) {
        this.resource = resource;
    }
    
    public String getResource() {
    	return this.resource;
    }
    
    /**
     * Liefert ein Material anhand seines Namens, oder <tt>null</tt>, falls es
     * nicht existiert.
     * @param materialName Der Name des Materials
     * @return Das Material
     */
    public Material getMaterial(String materialName) {
        return this.materials.get(materialName);
    }
    
    /**
     * Laedt die Daten aus der Resource.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init() {
        ResHandle handle = GameApp.getResCache().getHandle(new Resource(this.resource));
        if(handle != null) {
            boolean success = false;
            BufferedReader reader = new BufferedReader(new InputStreamReader(handle.getBufferAsInputStream()));
            try {
                success = this.loadData(reader);
                reader.close();
            } catch (IOException ex) {
                Logger.INSTANCE.error(ex.getMessage());
            }
            handle.unlock();
            return success;
        } else {
            return false;
        }
    }
    
    /**
     * Zerstoert alle Materialien.
     */
    public void destroy() {
        for(Material material : this.materials.values()) {
            material.delRef();
        }
        this.materials.clear();
    }
    
    /**
     * Laedt die Daten und pflegt sie in die Datenstruktur ein.
     * @param reader Die Daten
     * @return <tt>true</tt>, falls erfolgreich
     * @throws IOException Fehler des BufferedReader werden weitergereicht
     */
    private boolean loadData(BufferedReader reader) throws IOException {
        this.currentLine = 0;
        String line;
        this.currentMaterial = null;
        while((line = reader.readLine()) != null) {
            ++this.currentLine;
            line = line.trim();
            if(!line.isEmpty() && !line.startsWith("#")) {
                boolean success = true;
                StringTokenizer tokenizer = new StringTokenizer(line);
                String type = tokenizer.nextToken();
                if(this.currentMaterial == null && !type.equals("newmtl")) {
                    this.error("no current material");
                }
                String textureName = null;
                if(type.startsWith("map_")) {
                    String split[] = line.split(" ", 2);
                    if(split.length == 2) {
                        textureName = split[1];
                    } else {
                        this.error("illegal texture definition");
                        success = false;
                    }
                }
                switch(type) {
                    case "newmtl":
                        success = this.createMaterial(tokenizer);
                        break;
                    case "Ka":
                        success = this.setColor(this.currentMaterial.getKa(), tokenizer);
                        break;
                    case "Kd":
                        success = this.setColor(this.currentMaterial.getKd(), tokenizer);
                        break;
                    case "Ks":
                        success = this.setColor(this.currentMaterial.getKs(), tokenizer);
                        break;
                    case "map_d":
                        this.currentMaterial.setAlphaTextue(textureName);
                        break;
                    case "map_Ka":
                        this.currentMaterial.setAmbientTextue(textureName);
                        break;
                    case "map_Kd":                        
                        this.currentMaterial.setDiffuseTextue(textureName);
                        break;
                    case "map_Ks":                        
                        this.currentMaterial.setSpecularTextue(textureName);
                        break;
                    default:
                        //Logger.INSTANCE.warning(String.format("unsupported or illegal type \"%s\" was ignored (%s, %d)", type, this.resource, this.currentLine));
                        break;
                }
                if(!success) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Erzeugt ein neues Material.
     * @param data Die Daten des Materials
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean createMaterial(StringTokenizer data) {
        if(data.countTokens() != 1) {
            this.error("illegal newmtl line");
            return false;
        }
        String name = data.nextToken();
        if(this.materials.containsKey(name)) {
            this.error("material redefinition");
            return false;
        }
        
//        Logger.INSTANCE.error("Material Name: " + name);
        
        this.currentMaterial = new Material(name);
        this.materials.put(this.currentMaterial.getName(), this.currentMaterial);
        this.currentMaterial.addRef();
        return true;
    }
    
    /**
     * Liest eine Farbe aus einem String.
     * @param dest Der Zielvektor
     * @param data Die Daten der Farbe
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean setColor(Vector4f dest, StringTokenizer data) {
        if(data.countTokens() != 3 && data.countTokens() != 4) {
            this.error("wrong data count");
        }
        try {
            dest.x = Float.parseFloat(data.nextToken());
            dest.y = Float.parseFloat(data.nextToken());
            dest.z = Float.parseFloat(data.nextToken());
            if (data.hasMoreTokens()) {
                dest.w = Float.parseFloat(data.nextToken());
            } else {
                dest.w = 1.0f;
            }
        } catch (NumberFormatException ex) {
            this.error("illegal float expression");
            return false;            
        }
        return true;
    }
    
    
    /**
     * Loggt einen Fehler und fuegt den Namen und die Zeile der obj-Datei an.
     * @param message Die Fehlermeldung
     */
    private void error(String message) {
        Logger.INSTANCE.error(1, "%s (%s, %d)", message, this.resource, this.currentLine);
    }
}
