/*
 * Copyright (C) 2013 nico3000 - All Rights Reserved
 */
package mesh.obj;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import logging.Logger;
import main.GameApp;
import mesh.Material;
import org.lwjgl.util.vector.Vector2f;
import org.lwjgl.util.vector.Vector3f;
import resource.ResHandle;
import resource.Resource;

/**
 * Haelt die Daten einer Wavefront obj-Datei.
 * @author nico3000
 */
public class ObjData {
    /**
     * Name der Resource.
     */
    private final String resource;

	/**
     * Vertices der Datei.
     */
    private final List<Vector3f> vertexList = new ArrayList<>();
    /**
     * Normalen der Datei.
     */
    private final List<Vector3f> normalList = new ArrayList<>();
    /**
     * Texturkoordinaten der Datei.
     */
    private final List<Vector2f> texList = new ArrayList<>();
    /**
     * Faces der Datei.
     */
    private final Map<Material, List<Face>> faces = new HashMap<>();
    /**
     * Enthaelt waehrend des Ladens die aktuelle Zeile.
     */
    private int currentLine;
    /**
     * Enthaelt waehrend des Ladens die aktuelle Gruppe.
     */
    private String currentGroup = null;
    /**
     * Die aktuelle Materialbibliothek.
     */
    private MtlData currentMaterialLibrary = null;
    /**
     * Das aktuelle Material.
     */
    private Material currentMaterial = null;
    /**
     * Alle referenzierten Materialbibliotheken.
     */
    private final Map<String, MtlData> materialLibraries = new HashMap<>();
    
    /**
     * Konstruktor.
     * @param resource Der Name der Resource
     */
    public ObjData(String resource) {
        this.resource = resource;
    }
    
    /**
     * Laedt die Datei und pflegt den Inhalt in diese Instanz ein. Die Resource
     * wird im Anschluss wieder freigegeben.
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean init() {
        ResHandle handle = GameApp.getResCache().getHandle(new Resource(this.resource));
        if(handle != null) {
            boolean success = false;
            BufferedReader reader = new BufferedReader(new InputStreamReader(handle.getBufferAsInputStream()));
            try {
                success = this.loadData(reader);
                reader.close();
            } catch (IOException ex) {
                Logger.INSTANCE.error(ex.getMessage());
            }
            handle.unlock();
            return success;
        } else {
            return false;
        }
    }
    
    /**
     * Zerstoert alle gesammelten Daten, inklusive der Materialbibliotheken.
     */
    public void destroy() {
        for(MtlData data : this.materialLibraries.values()) {
            //data.destroy();
        }
    }
    
    /**
     * Liefert die in der obj-Datei verwendeten Materialien.
     * @return Die Materialien
     */
    public Set<Material> getMaterials() {
        return Collections.unmodifiableSet(this.faces.keySet());
    }
    
    //TODO Doku
    public Map<String, MtlData> getMaterialLibraries() {
    	return this.materialLibraries;
    }
    
    /**
     * Liefert zu einem Material die Liste aller Faces, die es benutzen.
     * @param material Das Material
     * @return Die Liste der Faces
     */
    public List<Face> getFacesByMaterial(Material material) {
        return Collections.unmodifiableList(this.faces.get(material));
    }
    
    /**
     * Extrahiert die Daten aus einem BufferedReader.
     * @param reader Der Inhalt der Datei
     * @return <tt>true</tt>, falls erfolgreich
     * @throws IOException Fehler des BufferedReader werden weitergereicht
     */
    private boolean loadData(BufferedReader reader) throws IOException {
        this.vertexList.clear();
        this.texList.clear();
        this.normalList.clear();
        this.faces.clear();
        this.faces.put(null, new LinkedList<Face>());
        String line;
        this.currentLine = 0;
        while((line = reader.readLine()) != null) {
            ++this.currentLine;
            line = line.trim();
            if(!line.isEmpty() && !line.startsWith("#")) {
                StringTokenizer tokenizer = new StringTokenizer(line.trim());
                String type = tokenizer.nextToken();
                boolean success;
                switch(type) {
                    case "v":
                        success = this.addVector3f(tokenizer, this.vertexList);
                        break;
                    case "vn":
                        success = this.addVector3f(tokenizer, this.normalList);
                        break;
                    case "vt":
                        success = this.addVector2f(tokenizer, this.texList);
                        break;
                    case "f":
                        success = this.addFace(tokenizer);
                        break;
                    case "g":
                        success = this.setGroup(tokenizer);
                        break;
                    case "usemtl":
                        success = this.setMaterial(tokenizer);
                        break;
                    case "mtllib":
                        String split[] = line.split(" ", 2);
                        if(split.length != 2) {
                            this.error("illegal mtllib definition");
                            success = false;
                        } else {
                            success = this.setMtlLib(split[1]);
                        }
                        break;
                    default:
                        Logger.INSTANCE.warning(String.format("unsupported or illegal type \"%s\" was ignored (%s, %d)", type, this.resource, this.currentLine));
                        success = true;
                        break;
                }
                if(!success) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * Setzt die Materialbibliothek.
     * @param data Der Dateiname der Bibliothek
     * @return <tt>true</tt>, falls erfolgreich
     */
    public boolean setMtlLib(String library) {
    	boolean success = false;
    	
    	String libraryName = library.substring(7, library.length() - 4);
    	String resourceName = this.resource.substring(7, this.resource.length() - 4);

    	MtlData materialLibrary = this.materialLibraries.get(library);
        if (materialLibrary == null) {
        	materialLibrary = new MtlData(library);
        	//TODO es werden eventuell auch Texturen von customLibraries geladen, die garnicht genutzt werden
            success = materialLibrary.init();
            if(success) {
                this.materialLibraries.put(library, materialLibrary);
                if (libraryName.equals(resourceName) || this.currentMaterialLibrary == null) {
                	this.currentMaterialLibrary = materialLibrary;
                }
            }
        }
        return success;
    }
    
    /**
     * Setzt die aktuelle Gruppe.
     * @param data Der Name der Gruppe
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean setGroup(StringTokenizer data) {
        if(data.countTokens() != 1) {
            this.currentGroup = null;
        } else {
            this.currentGroup = data.nextToken();
        }
        return true;
    }
    
    /**
     * Setzt das aktuelle Material.
     * @param data Der Name des Materials
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean setMaterial(StringTokenizer data) {
        if(data.countTokens() != 1) {
            this.error("illegal material usage");
            return false;            
        }
        String name = data.nextToken();
        if(this.currentMaterialLibrary == null) {
            this.error("no current material library to load material from");
            return false;
        }
        this.currentMaterial = this.currentMaterialLibrary.getMaterial(name);
        if(!this.faces.containsKey(this.currentMaterial)) {
            this.faces.put(this.currentMaterial, new LinkedList<Face>());
        }
        if(this.currentMaterial == null) {
            this.error("unknown material \"" + name + "\"");
            return false;
        }
        return true;
    }
    
    /**
     * Wandelt die Daten eines Strings in einen Vector3f um und fuegt ihn bei
     * Erfolg einer Liste hinzu.
     * @param data Die Daten
     * @param list Die Liste
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean addVector3f(StringTokenizer data, List<Vector3f> list) {
        if(data.countTokens() < 3) {
            this.error("wrong data count");
            return false;
        }
        float x, y, z;
        try {
            x = Float.parseFloat(data.nextToken());
            y = Float.parseFloat(data.nextToken());
            z = Float.parseFloat(data.nextToken());
        } catch (NumberFormatException ex) {
            this.error("illegal float expression");
            return false;
        }        
        list.add(new Vector3f(x, y, z));
        return true;
    }
    
    /**
     * Wandelt die Daten eines Strings in einen Vector2f um und fuegt ihn bei
     * Erfolg einer Liste hinzu.
     * @param data Die Daten
     * @param list Die Liste
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean addVector2f(StringTokenizer data, List<Vector2f> list) {
        if(data.countTokens() < 2) {
            this.error("wrong data count");
            return false;
        }
        float x, y;
        try {
            x = Float.parseFloat(data.nextToken());
            y = Float.parseFloat(data.nextToken());
        } catch (NumberFormatException ex) {
            this.error("illegal float expression");
            return false;
        }
        list.add(new Vector2f(x, y));
        return true;
    }
    
    /**
     * Wandelt die Daten eines Strings in ein Face um, wie es die
     * obj-Spezifikation vorsieht.
     * @param data Die Daten
     * @return <tt>true</tt>, falls erfolgreich
     */
    private boolean addFace(StringTokenizer data) {
        int count = data.countTokens();
        if(count < 2) {
            this.error("illegal face");
            return false;
        }
        Face face = new Face(count);
        for(int i=0; i < count; ++i) {
            String split[] = data.nextToken().split("/");
            FaceType type = this.getFaceType(split);
            if(type == FaceType.UNKOWN) {
                this.error("illegal face index " + i);
                return false;
            }
            if(face.type == FaceType.UNKOWN) {
                face.type = type;
            } else if(face.type != type) {
                this.error("face with different index types");
                return false;
            }
            try {
                face.setVertexIndex(i, Integer.parseInt(split[0]));
                switch (face.type) {
                    case VERTEX_TEXTURE_NORMAL:
                        face.setNormalIndex(i, Integer.parseInt(split[2]));     
                        // no break
                    case VERTEX_TEXTURE:
                        face.setTexIndex(i, Integer.parseInt(split[1]));
                        break;
                    case VERTEX_NORMAL:
                        face.setNormalIndex(i, Integer.parseInt(split[2]));
                        break;
                }
            } catch (NumberFormatException ex) {
                this.error("illegal int expression");
                return false;
            }
        }
        this.faces.get(this.currentMaterial).add(face);
        return true;
    }
    
    /**
     * Bestimmt aus einem an <tt>/</tt> geteilten Strings die Art des Faces.
     * @param split Der an <tt>t</tt> geteilte String
     * @return Der Typ des Faces
     */
    private FaceType getFaceType(String split[]) {
        if(split.length == 1) {
            return FaceType.VERTEX;
        } else if(split.length == 2) {
            return FaceType.VERTEX_TEXTURE;
        } else if(split.length == 3) {
            if(split[1].isEmpty()) {
                return FaceType.VERTEX_NORMAL;
            } else {
                return FaceType.VERTEX_TEXTURE_NORMAL;
            }
        }
        return FaceType.UNKOWN;
    }
    
    /**
     * Loggt einen Fehler und fuegt den Namen und die Zeile der obj-Datei an.
     * @param message Die Fehlermeldung
     */
    private void error(String message) {
        Logger.INSTANCE.error(1, "%s (%s, %d)", message, this.resource, this.currentLine);
    }
    
    public Vector3f getVertex(int i) {
        return this.vertexList.get(i);
    }
    
    public Vector3f getNormal(int i) {
        return i < this.normalList.size() ? this.normalList.get(i) : new Vector3f();
    }
    
    public Vector2f getTexture(int i) {
        return i < this.texList.size() ? this.texList.get(i) : new Vector2f();
    }
    
    public String getResource() {
		return resource;
	}
    
    /**
     * Moegliche Typen von Faces in einer obj-Datei.
     */
    public enum FaceType {
        /**
         * Unbekannt / Unegueltig.
         */
        UNKOWN,
        /**
         * Nur Vertices.
         */
        VERTEX,
        /**
         * Vertices und Texturkoordinaten.
         */
        VERTEX_TEXTURE,
        /**
         * Vertices und Normalen.
         */
        VERTEX_NORMAL,
        /**
         * Vertices, Texturkoordinaten und Normalen.
         */
        VERTEX_TEXTURE_NORMAL,
    }
    
    public class ObjIndex {
        private int vertex;
        private int normal;
        private int tex;

        private ObjIndex(int vertex, int normal, int tex) {
            this.vertex = vertex;
            this.normal = normal;
            this.tex = tex;
        }

        public int getVertex() {
            return vertex;
        }

        public int getNormal() {
            return normal;
        }

        public int getTex() {
            return tex;
        }

        @Override
        public int hashCode() {
            int hash = 3;
            hash = 61 * hash + this.vertex;
            hash = 61 * hash + this.normal;
            hash = 61 * hash + this.tex;
            return hash;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final ObjIndex other = (ObjIndex) obj;
            if (this.vertex != other.vertex) {
                return false;
            }
            if (this.normal != other.normal) {
                return false;
            }
            if (this.tex != other.tex) {
                return false;
            }
            return true;
        }
    }
    
    /**
     * Sammelt die Indizes eines Faces.
     */
    public class Face {
        /**
         * Die Indizes des Faces.
         */
        private final ObjIndex indices[];
        /**
         * Der Typ des Faces.
         */
        private FaceType type = FaceType.UNKOWN;
        /**
         * Gruppe des Faces.
         */
        private final String group;
        /**
         * Materialname des Faces.
         */
        private final Material material;
        
        /**
         * Konstruktor.
         * @param indexCount Die Anzahl der Indizes
         */
        private Face(int indexCount) {
            this.indices = new ObjIndex[indexCount];
            this.group = ObjData.this.currentGroup;
            this.material = ObjData.this.currentMaterial;
            for(int i=0; i < indexCount; ++i) {
                this.indices[i] = new ObjIndex(0, 0, 0);
            }
        }
        
        /**
         * Fuegt den Index eines Vertices dem Face hinzu. Ist der Index negativ
         * (stellt also eine relative Indizierung dar), wird dieser in einen
         * absoluten Index ueberfuehrt.
         * @param pos Die Nummer des Indizes
         * @param vertexIndex Der Index
         */
        private void setVertexIndex(int pos, int vertexIndex) {
            if(vertexIndex < 0) {
                this.indices[pos].vertex = ObjData.this.vertexList.size() + vertexIndex;
            } else {
                this.indices[pos].vertex = vertexIndex - 1;
            }
        }
        
        /**
         * Fuegt den Index einer Texturkoordinate dem Face hinzu. Ist der Index
         * negativ (stellt also eine relative Indizierung dar), wird dieser in
         * einen absoluten Index ueberfuehrt.
         * @param pos Die Nummer des Indizes
         * @param texIndex Der Index
         */
        private void setTexIndex(int pos, int texIndex) {
            if(texIndex < 0) {
                this.indices[pos].tex = ObjData.this.texList.size() + texIndex;
            } else {
                this.indices[pos].tex = texIndex - 1;
            }
        }
        
        /**
         * Fuegt den Index einer Normalen dem Face hinzu. Ist der Index negativ
         * (stellt also eine relative Indizierung dar), wird dieser in einen
         * absoluten Index ueberfuehrt.
         * @param pos Die Nummer des Indizes
         * @param normalIndex Der Index
         */        
        private void setNormalIndex(int pos, int normalIndex) {
            if(normalIndex < 0) {
                this.indices[pos].normal = ObjData.this.normalList.size() + normalIndex;
            } else {
                this.indices[pos].normal = normalIndex - 1;
            }
        }

        public FaceType getType() {
            return this.type;
        }
        
        public int getIndexCount() {
            return this.indices.length;
        }
        
        public ObjIndex getIndex(int i) {
            return this.indices[i];
        }
    }
}
