/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package process;

import util.MathUtil;

/**
 * Basisklasse eines Prozesses.
 * @author nico3000
 */
public abstract class EngineProcess {    
    /**
     * Moegliche Stati eines Prozesses.
     */
    static enum State {
        /**
         * Ein Prozess, der gerade erstellt wurde.
         */
        UNINITIALIZED,
        /**
         * Ein Prozess, der dem System nicht mehr bekannt ist.
         */
        REMOVED,
        /**
         * Ein gerade laufender Prozess.
         */
        RUNNING,
        /**
         * Ein pausierter Prozess.
         */
        PAUSED,
        /**
         * Ein erfolgreich beendeter Prozess. Wird im naechsten Durchgang
         * entfernt.
         */
        SUCCEEDED,
        /**
         * Ein fehlgeschlagener Prozess. Wird im naechsten Durchgang entfernt.
         */
        FAILED,
        /**
         * Ein abgebrochener Prozess. Wird im naechsten Durchgang entfernt.
         */
        ABORTED,
    }
    
    /**
     * Aktueller Fortschritt des Prozesses.
     */
    private float progress = 0.0f;    
    
    /**
     * Aktueller Status des Prozesses.
     */
    private State state = State.UNINITIALIZED;
    
    /**
     * Moeglicher Folgeprozess.
     */
    private EngineProcess child = null;
    
    /**
     * Setzt den Status des Prozesses.
     * @param state Der neue Status
     */
    final void setState(State state) {
        this.state = state;
    }
    
    /**
     * Liefert den aktuellen Status des Prozesses zurueck.
     * @return Der aktuelle Status
     */
    final State getState() {
        return state;
    }
    
    /**
     * Setzt den aktuellen Fortschritt des Prozesses. Der Wert wird zwischen 0
     * und 1 geclampt.
     * @param progress Der Fortschritt 
     */
    protected final void setProgress(float progress) {
        this.progress = MathUtil.clamp(progress, 0.0f, 1.0f);
    }
    
    /**
     * Setzt einen direkten Nachfolgeprozess, der nach dem Beenden dieses
     * Prozesses gestartet wird.
     * @param child Der Nachfolgeprozess
     */
    public final void attachChild(EngineProcess child) {
        this.child = child;
    }
    
    /**
     * Entfernt den direkten Nachfolgeprozess.
     */
    public final void removeChild() {
        this.attachChild(null);
    }
    
    /**
     * Liefert den direkten Nachfolgeprozess.
     * @return Der Nachfolgeprozess
     */
    final EngineProcess getChild() {
        return this.child;
    }
    
    /**
     * Checkt, ob der Prozess noch aktiv (oder pausiert) ist.
     * @return <tt>true</tt>, falls aktiv oder pausiert
     */
    public final boolean isAlive() {
        return this.state == State.PAUSED || this.state == State.RUNNING;
    }
    
    /**
     * Checkt, ob der Prozess bereits beendet wurde.
     * @return <tt>true</tt>, falls Prozess abgeschlossen, abgebrochen oder gescheitert
     */
    public final boolean isDead() {
        return this.state == State.ABORTED || this.state == State.FAILED || this.state == State.SUCCEEDED;
    }
    
    /**
     * Checkt, ob der Prozess bereits entfernt wurde.
     * @return <tt>true</tt>, falls der Prozess entfernt wurde
     */
    public final boolean isRemoved() {
        return this.state == State.REMOVED;
    }
    
    /**
     * Checkt, ob der Prozess pausiert ist.
     * @return <tt>true</tt>, falls pausiert
     */
    public final boolean isPaused() {
        return this.state == State.PAUSED;
    }
    
    /**
     * Setzt den Status auf <i>erfolgreich beendet</i>.
     */
    public final void succeed() {
        if(this.isAlive()) {
            this.state = State.SUCCEEDED;
        }
    }
    
    /**
     * Setzt den Status auf <i>gescheitert</i>.
     */
    public final void fail() {
        if(this.isAlive()) {
            this.state = State.FAILED;
        }
    }
    
    /**
     * Pausiert den Prozess.
     */
    public final void pause() {
        if(this.isAlive()) {
            this.state = State.PAUSED;
        }
    }
    
    /**
     * Beendet die Pause eines Prozesses.
     */
    public final void resume() {
        if(this.isPaused()) {
            this.state = State.RUNNING;
        }
    }
    
    /**
     * Wird aufgerufen, sobald der Prozess initialisert wurde.
     */
    protected void onInit() {
        this.state = State.RUNNING;
    }    
    
    /**
     * Wird aufgerufen, sobald der Prozess erfolgreich beendet wurde.
     */
    protected void onSuccess() {}
    
    /**
     * Wird aufgerufen, sobald der Prozess vorzeitig abgebrochen wurde.
     */
    protected void onAbort() {}
    
    /**
     * Wird aufgerufen, sobald der Prozess gescheitert ist.
     */
    protected void onFail() {}
    
    /**
     * Wird fuer einen aktiven Prozess jeden Frame aufgerufen.
     * @param dMillis Die Anzahl der Millisekunden, die seit dem letzten Aufruf
     *                vergangen sind
     */
    protected abstract void onUpdate(long dMillis);
    
    /**
     * Getter fuer den Fortschritt des Prozesses.
     * @return Der aktuelle Fortschritt, liegt zwischen 0 und 1
     */
    public float getProgress() { 
        return this.progress;
    }
    
    
}
