/*
 * Copyright (C) 2012 nico3000 - All Rights Reserved
 */
package view.scene.nodes;

import actor.ActorFactory;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import logging.Logger;
import view.scene.SceneScreenElement;

/**
 * Basisklasse fuer einen Knoten des Szenegraphen.
 * @author nico3000
 */
public abstract class SceneNode {
    /**
     * Mit diesem Knoten verknuepfter Actor.
     */
    private int actorId = ActorFactory.INVALID_ACTOR_ID;
    /**
     * Der Vaterknoten dieses Knotens.
     */
    private SceneNode parent = null;
    /**
     * Liste der Kindknoten.
     */
    private final List<SceneNode> children = new LinkedList<>();
    /**
     * Gibt an, ob Kindknoten erlaubt sind.
     */
    protected boolean allowChildren = true;
    /**
     * Gibt an, ob sich der zum Knoten gehoerige Actor oder einer seiner
     * Vorfahren bewegt hat.
     */
    protected boolean hasMoved = true;

    /**
     * Setzt den mit dem Knoten verknuepften Actor.
     * @param actorId Die ID des Actors
     */
    public void setActorId(int actorId) {
        this.actorId = actorId;
    }

    /**
     * Liefert die ID des mit diesem Knoten verknuepften Actors.
     * @return Die ID des Actors
     */
    public final int getActorId() {
        return actorId;
    }
    
    /**
     * Wird jeden Frame aufgerufen und aktualisiert den Zustand des Knotens.
     * @param scene Die Szene, zu der der Knoten gehoert
     * @param deltaMillis Dauer des Frames in Millisekunden (Gametimer)
     */
    public void update(SceneScreenElement scene, long deltaMillis) {
        for(SceneNode child : this.children) {
            child.update(scene, deltaMillis);
        }
    }
    
    /**
     * Wird aufgerufen, falls das Graphicsdevice verloren geht.
     */
    public void lostDevice() {
        for(SceneNode child : this.children) {
            child.lostDevice();
        }
    }
    
    /**
     * Wird initial einmal und nach jedem Wiederherstellen des Graphicsdevice
     * aufgerufen.
     */
    public void restore() {
        for(SceneNode child : this.children) {
            child.restore();
        }        
    }
    
    /**
     * Wird aufgerufen, wenn der Knoten zerstoert werden soll.
     */
    public void destroy() {
        this.parent = null;
        for(SceneNode child : this.children) {
            child.destroy();
        }
        //Logger.INSTANCE.info("SceneNode of actor " + this.getActorId() + " destroyed");
    }
    
    /**
     * Fuegt dem Knoten einen Kindknoten hinzu. Die folgenden Bedingungen
     * muessen alle erfuellt sein.
     * <ul>
     * <li>Der Knoten erlaubt es</li>
     * <li>Der uebergebene Knoten ist nicht dieser Knoten selbst</li>
     * <li>Der uebergebene Knoten ist nicht bereits ein Kindknoten</li>
     * </ul>
     * @param child Der neue Kindknoten
     * @return <tt>true</tt>, falls erfolgreich
     */
    public final boolean addChild(SceneNode child) {
        if(!this.allowChildren) {
            Logger.INSTANCE.error("tried to add a node to a leaf");
            return false;
        } else if(child == this) {
            Logger.INSTANCE.error("tried to add a node as a child of itself");
            return false;
        } else if(child.parent != null) {
            Logger.INSTANCE.error("tried to add a node twice");
            return false;            
        } else {
            child.parent = this;
            this.children.add(child);
            return true;
        }
    }
    
    /**
     * Entfernt einen Nachfahren dieses Knotens anhand einer ActorID.
     * @param actorId Die ID des Actors
     * @return <tt>true</tt>, falls erfolgreich
     */
    public final boolean removeChild(int actorId) {
        if(!this.allowChildren) {
            return false;
        } else {
            ListIterator<SceneNode> iter = this.children.listIterator();
            while(iter.hasNext()) {
                SceneNode child = iter.next();
                if(child.getActorId() == actorId) {
                    iter.remove();
                    child.destroy();
                    return true;
                }
            }
            iter = this.children.listIterator();
            while(iter.hasNext()) {
                SceneNode child = iter.next();
                if(child.removeChild(actorId)) {
                    return true;
                }
            }
        }
        return false;
    }
    
    /**
     * Durchsucht den Knoten und seine Nachfahren nach einem Knoten, der einer
     * gegebenen ActorID zugeordnet ist.
     * @param actorId Die ID des Actors
     * @return Der mit der ID verknuepfte Knoten oder <tt>null</tt>, wenn nicht
     *         vorhanden
     */
    public final SceneNode getNode(int actorId) {
        if(this.getActorId() == actorId) {
            return this;
        } else {
            SceneNode toReturn;
            for(SceneNode child : this.children) {
                if((toReturn = child.getNode(actorId)) != null) {
                    return toReturn;
                }
            }
            return null;
        }
    }
    
    /**
     * Prueft, ob der Node bzgl seiner Szene sichtbar ist.
     * @return <tt>true</tt>, falls sichtbar
     */
    public abstract boolean isVisible(SceneScreenElement scene);
    
    /**
     * Bereitet das Rendern des Knoten vor.
     * @param scene Die Szene des Knotens
     */
    public abstract void preRender(SceneScreenElement scene);
    
    /**
     * Rendert den Knoten.
     * @param scene Die Szene des Knotens
     */
    public abstract void render(SceneScreenElement scene);
    
    /**
     * Rendert die Kinder des Knotens, sofern vorhanden
     * @param scene Die Szene des Knotens
     */
    public final void renderChildren(SceneScreenElement scene) {
        if(this.allowChildren) {
            for(SceneNode child : this.children) {
                child.preRender(scene);
                if(child.isVisible(scene)) {                    
                    child.render(scene);
                    child.renderChildren(scene);
                }
                child.postRender(scene);
            }
        }
    }
    
    /**
     * Schliesst das Rendern ab und raeumt auf.
     * @param scene Die Szene des Knotens
     */
    public abstract void postRender(SceneScreenElement scene);
    
    /**
     * Signalisiert, dass sich dieser Knoten bewegt hat.
     */
    public void actorMoved() {
        this.predecessorMoved();
        this.successorMoved();
    }
    
    /**
     * Signalisiert, dass ein Nachfahre oder der Knoten selbst sich bewegt hat.
     */
    protected void successorMoved() {
        if(this.parent != null) {
            this.parent.successorMoved();
        }
    }
    
    /**
     * Signalisiert, dass ein Vorfahre oder der Knoten selbst sich bewegt hat.
     */
    protected void predecessorMoved() {
        this.hasMoved = true;
        if(this.allowChildren) {
            for(SceneNode child : this.children) {
                child.predecessorMoved();
            }
        }
    }

    /**
     * Liefert eine nicht veraenderbare Liste der Kinder dieses Knotens.
     * @return Die Liste der Kinder
     */
    protected List<SceneNode> getChildren() {
        return Collections.unmodifiableList(this.children);
    }
}
